"""Base Indy Verifier class."""

import logging
from abc import ABC, ABCMeta, abstractmethod
from enum import Enum
from time import time
from typing import Mapping, Tuple

from ..core.profile import Profile
from ..ledger.multiple_ledger.ledger_requests_executor import (
    GET_CRED_DEF,
    IndyLedgerRequestsExecutor,
)
from ..messaging.util import canon, encode
from ..multitenant.base import BaseMultitenantManager
from .models.xform import indy_proof_req2non_revoc_intervals

LOGGER = logging.getLogger(__name__)


class PresVerifyMsg(str, Enum):
    """Credential verification codes."""

    RMV_REFERENT_NON_REVOC_INTERVAL = "RMV_RFNT_NRI"
    RMV_GLOBAL_NON_REVOC_INTERVAL = "RMV_GLB_NRI"
    TSTMP_OUT_NON_REVOC_INTRVAL = "TS_OUT_NRI"
    CT_UNREVEALED_ATTRIBUTES = "UNRVL_ATTR"
    PRES_VALUE_ERROR = "VALUE_ERROR"
    PRES_VERIFY_ERROR = "VERIFY_ERROR"


class IndyVerifier(ABC, metaclass=ABCMeta):
    """Base class for Indy Verifier."""

    def __repr__(self) -> str:
        """Return a human readable representation of this class.

        Returns:
            A human readable string for this class

        """
        return "<{}>".format(self.__class__.__name__)

    def non_revoc_intervals(self, pres_req: dict, pres: dict, cred_defs: dict) -> list:
        """Remove superfluous non-revocation intervals in presentation request.

        Irrevocable credentials constitute proof of non-revocation, but
        indy rejects proof requests with non-revocation intervals lining up
        with non-revocable credentials in proof: seek and remove.

        Args:
            pres_req: presentation request
            pres: corresponding presentation
            cred_defs: credential definitions by cred def id

        """
        msgs = []
        for req_proof_key, pres_key in {
            "revealed_attrs": "requested_attributes",
            "revealed_attr_groups": "requested_attributes",
            "predicates": "requested_predicates",
        }.items():
            for uuid, spec in pres["requested_proof"].get(req_proof_key, {}).items():
                if (
                    "revocation"
                    not in cred_defs[
                        pres["identifiers"][spec["sub_proof_index"]]["cred_def_id"]
                    ]["value"]
                ):
                    if uuid in pres_req[pres_key] and pres_req[pres_key][uuid].pop(
                        "non_revoked", None
                    ):
                        msgs.append(
                            f"{PresVerifyMsg.RMV_REFERENT_NON_REVOC_INTERVAL.value}::"
                            f"{uuid}"
                        )
                        LOGGER.info(
                            (
                                "Amended presentation request (nonce=%s): removed "
                                "non-revocation interval at %s referent "
                                "%s; corresponding credential in proof is irrevocable"
                            ),
                            pres_req["nonce"],
                            pres_key,
                            uuid,
                        )

        if all(
            (
                spec.get("timestamp") is None
                and "revocation" not in cred_defs[spec["cred_def_id"]]["value"]
            )
            for spec in pres["identifiers"]
        ):
            pres_req.pop("non_revoked", None)
            msgs.append(PresVerifyMsg.RMV_GLOBAL_NON_REVOC_INTERVAL.value)
            LOGGER.warning(
                (
                    "Amended presentation request (nonce=%s); removed global "
                    "non-revocation interval; no revocable credentials in proof"
                ),
                pres_req["nonce"],
            )
        return msgs

    async def check_timestamps(
        self,
        profile: Profile,
        pres_req: Mapping,
        pres: Mapping,
        rev_reg_defs: Mapping,
    ) -> list:
        """Check for suspicious, missing, and superfluous timestamps.

        Raises ValueError on timestamp in the future, prior to rev reg creation,
        superfluous or missing.

        Args:
            profile: profile
            pres_req: indy proof request
            pres: indy proof request
            rev_reg_defs: rev reg defs by rev reg id, augmented with transaction times
        """
        msgs = []
        now = int(time())
        non_revoc_intervals = indy_proof_req2non_revoc_intervals(pres_req)
        LOGGER.debug(f">>> got non-revoc intervals: {non_revoc_intervals}")
        # timestamp for irrevocable credential
        cred_defs = []
        for index, ident in enumerate(pres["identifiers"]):
            LOGGER.debug(f">>> got (index, ident): ({index},{ident})")
            cred_def_id = ident["cred_def_id"]
            multitenant_mgr = profile.inject_or(BaseMultitenantManager)
            if multitenant_mgr:
                ledger_exec_inst = IndyLedgerRequestsExecutor(profile)
            else:
                ledger_exec_inst = profile.inject(IndyLedgerRequestsExecutor)
            ledger = (
                await ledger_exec_inst.get_ledger_for_identifier(
                    cred_def_id,
                    txn_record_type=GET_CRED_DEF,
                )
            )[1]
            async with ledger:
                cred_def = await ledger.get_credential_definition(cred_def_id)
            cred_defs.append(cred_def)
            if ident.get("timestamp"):
                if not cred_def["value"].get("revocation"):
                    raise ValueError(
                        f"Timestamp in presentation identifier #{index} "
                        f"for irrevocable cred def id {cred_def_id}"
                    )

        # timestamp in the future too far in the past
        for ident in pres["identifiers"]:
            timestamp = ident.get("timestamp")
            rev_reg_id = ident.get("rev_reg_id")

            if not timestamp:
                continue

            if timestamp > now + 300:  # allow 5 min for clock skew
                raise ValueError(f"Timestamp {timestamp} is in the future")
            reg_def = rev_reg_defs.get(rev_reg_id)
            if not reg_def:
                raise ValueError(f"Missing registry definition for '{rev_reg_id}'")
            if "txnTime" not in reg_def:
                raise ValueError(
                    f"Missing txnTime for registry definition '{rev_reg_id}'"
                )
            if timestamp < reg_def["txnTime"]:
                raise ValueError(
                    f"Timestamp {timestamp} predates rev reg {rev_reg_id} creation"
                )

        # timestamp superfluous, missing, or outside non-revocation interval
        revealed_attrs = pres["requested_proof"].get("revealed_attrs", {})
        unrevealed_attrs = pres["requested_proof"].get("unrevealed_attrs", {})
        revealed_groups = pres["requested_proof"].get("revealed_attr_groups", {})
        self_attested = pres["requested_proof"].get("self_attested_attrs", {})
        preds = pres["requested_proof"].get("predicates", {})
        for uuid, req_attr in pres_req["requested_attributes"].items():
            if "name" in req_attr:
                if uuid in revealed_attrs:
                    index = revealed_attrs[uuid]["sub_proof_index"]
                    if cred_defs[index]["value"].get("revocation"):
                        timestamp = pres["identifiers"][index].get("timestamp")
                        if (timestamp is not None) ^ bool(non_revoc_intervals.get(uuid)):
                            LOGGER.debug(f">>> uuid: {uuid}")
                            LOGGER.debug(
                                f">>> revealed_attrs[uuid]: {revealed_attrs[uuid]}"
                            )
                            raise ValueError(
                                f"Timestamp on sub-proof #{index} "
                                f"is {'superfluous' if timestamp else 'missing'} "
                                f"vs. requested attribute {uuid}"
                            )
                        if non_revoc_intervals.get(uuid) and not (
                            non_revoc_intervals[uuid].get("from", 0)
                            < timestamp
                            < non_revoc_intervals[uuid].get("to", now)
                        ):
                            msgs.append(
                                f"{PresVerifyMsg.TSTMP_OUT_NON_REVOC_INTRVAL.value}::"
                                f"{uuid}"
                            )
                            LOGGER.info(
                                "Timestamp %s from ledger for item %s falls outside "
                                "non-revocation interval %s",
                                timestamp,
                                uuid,
                                non_revoc_intervals[uuid],
                            )
                elif uuid in unrevealed_attrs:
                    # nothing to do, attribute value is not revealed
                    msgs.append(f"{PresVerifyMsg.CT_UNREVEALED_ATTRIBUTES.value}::{uuid}")
                elif uuid not in self_attested:
                    raise ValueError(
                        f"Presentation attributes mismatch requested attribute {uuid}"
                    )

            elif "names" in req_attr:
                group_spec = revealed_groups.get(uuid)
                if (
                    group_spec is None
                    or "sub_proof_index" not in group_spec
                    or "values" not in group_spec
                ):
                    raise ValueError(f"Missing requested attribute group {uuid}")
                index = group_spec["sub_proof_index"]
                if cred_defs[index]["value"].get("revocation"):
                    timestamp = pres["identifiers"][index].get("timestamp")
                    if (timestamp is not None) ^ bool(non_revoc_intervals.get(uuid)):
                        raise ValueError(
                            f"Timestamp on sub-proof #{index} "
                            f"is {'superfluous' if timestamp else 'missing'} "
                            f"vs. requested attribute group {uuid}"
                        )
                    if non_revoc_intervals.get(uuid) and not (
                        non_revoc_intervals[uuid].get("from", 0)
                        < timestamp
                        < non_revoc_intervals[uuid].get("to", now)
                    ):
                        msgs.append(
                            f"{PresVerifyMsg.TSTMP_OUT_NON_REVOC_INTRVAL.value}::{uuid}"
                        )
                        LOGGER.info(
                            "Timestamp %s from ledger for item %s falls outside "
                            "non-revocation interval %s",
                            timestamp,
                            uuid,
                            non_revoc_intervals[uuid],
                        )

        for uuid, req_pred in pres_req["requested_predicates"].items():
            pred_spec = preds.get(uuid)
            if pred_spec is None or "sub_proof_index" not in pred_spec:
                raise ValueError(
                    f"Presentation predicates mismatch requested predicate {uuid}"
                )
            index = pred_spec["sub_proof_index"]
            if cred_defs[index]["value"].get("revocation"):
                timestamp = pres["identifiers"][index].get("timestamp")
                if (timestamp is not None) ^ bool(non_revoc_intervals.get(uuid)):
                    raise ValueError(
                        f"Timestamp on sub-proof #{index} "
                        f"is {'superfluous' if timestamp else 'missing'} "
                        f"vs. requested predicate {uuid}"
                    )
                if non_revoc_intervals.get(uuid) and not (
                    non_revoc_intervals[uuid].get("from", 0)
                    < timestamp
                    < non_revoc_intervals[uuid].get("to", now)
                ):
                    msgs.append(
                        f"{PresVerifyMsg.TSTMP_OUT_NON_REVOC_INTRVAL.value}::{uuid}"
                    )
                    LOGGER.warning(
                        f"Best-effort timestamp {timestamp} "
                        "from ledger falls outside non-revocation interval "
                        f"{non_revoc_intervals[uuid]}"
                    )
        return msgs

    async def pre_verify(self, pres_req: dict, pres: dict) -> list:
        """Check for essential components and tampering in presentation.

        Visit encoded attribute values against raw, and predicate bounds,
        in presentation, cross-reference against presentation request.

        Args:
            pres_req: presentation request
            pres: corresponding presentation

        """
        msgs = []
        if not (
            pres_req
            and "requested_predicates" in pres_req
            and "requested_attributes" in pres_req
        ):
            raise ValueError("Incomplete or missing proof request")
        if not pres:
            raise ValueError("No proof provided")
        if "requested_proof" not in pres:
            raise ValueError("Presentation missing 'requested_proof'")
        if "proof" not in pres:
            raise ValueError("Presentation missing 'proof'")

        for uuid, req_pred in pres_req["requested_predicates"].items():
            try:
                canon_attr = canon(req_pred["name"])
                matched = False
                found = False
                for ge_proof in pres["proof"]["proofs"][
                    pres["requested_proof"]["predicates"][uuid]["sub_proof_index"]
                ]["primary_proof"]["ge_proofs"]:
                    pred = ge_proof["predicate"]
                    if pred["attr_name"] == canon_attr:
                        found = True
                        if pred["value"] == req_pred["p_value"]:
                            matched = True
                            break
                if not matched:
                    raise ValueError(f"Predicate value != p_value: {pred['attr_name']}")
                    break
                elif not found:
                    raise ValueError(f"Missing requested predicate '{uuid}'")
            except (KeyError, TypeError):
                raise ValueError(f"Missing requested predicate '{uuid}'")

        revealed_attrs = pres["requested_proof"].get("revealed_attrs", {})
        unrevealed_attrs = pres["requested_proof"].get("unrevealed_attrs", {})
        revealed_groups = pres["requested_proof"].get("revealed_attr_groups", {})
        self_attested = pres["requested_proof"].get("self_attested_attrs", {})
        for uuid, req_attr in pres_req["requested_attributes"].items():
            if "name" in req_attr:
                if uuid in revealed_attrs:
                    pres_req_attr_spec = {req_attr["name"]: revealed_attrs[uuid]}
                elif uuid in unrevealed_attrs:
                    # unrevealed attribute, nothing to do
                    pres_req_attr_spec = {}
                    msgs.append(f"{PresVerifyMsg.CT_UNREVEALED_ATTRIBUTES.value}::{uuid}")
                elif uuid in self_attested:
                    if not req_attr.get("restrictions"):
                        continue
                    raise ValueError(
                        "Attribute with restrictions cannot be self-attested: "
                        f"'{req_attr['name']}'"
                    )
                else:
                    raise ValueError(f"Missing requested attribute '{req_attr['name']}'")
            elif "names" in req_attr:
                group_spec = revealed_groups[uuid]
                pres_req_attr_spec = {
                    attr: {
                        "sub_proof_index": group_spec["sub_proof_index"],
                        **group_spec["values"].get(attr),
                    }
                    for attr in req_attr["names"]
                }
            else:
                raise ValueError(
                    f"Request attribute missing 'name' and 'names': '{uuid}'"
                )

            for attr, spec in pres_req_attr_spec.items():
                try:
                    primary_enco = pres["proof"]["proofs"][spec["sub_proof_index"]][
                        "primary_proof"
                    ]["eq_proof"]["revealed_attrs"][canon(attr)]
                except (KeyError, TypeError):
                    raise ValueError(f"Missing revealed attribute: '{attr}'")
                if primary_enco != spec["encoded"]:
                    raise ValueError(f"Encoded representation mismatch for '{attr}'")
                if primary_enco != encode(spec["raw"]):
                    raise ValueError(f"Encoded representation mismatch for '{attr}'")
        return msgs

    @abstractmethod
    def verify_presentation(
        self,
        presentation_request,
        presentation,
        schemas,
        credential_definitions,
        rev_reg_defs,
        rev_reg_entries,
    ) -> Tuple[bool, list]:
        """Verify a presentation.

        Args:
            presentation_request: Presentation request data
            presentation: Presentation data
            schemas: Schema data
            credential_definitions: credential definition data
            rev_reg_defs: revocation registry definitions
            rev_reg_entries: revocation registry entries
        """



"""Verifiable Credential marshmallow schema classes."""

from datetime import datetime
from typing import List, Optional, Union

from dateutil import tz
from marshmallow import INCLUDE, ValidationError, fields, post_dump

from ....messaging.models.base import BaseModel, BaseModelSchema
from ....messaging.valid import (
    CREDENTIAL_CONTEXT_EXAMPLE,
    CREDENTIAL_CONTEXT_VALIDATE,
    CREDENTIAL_STATUS_EXAMPLE,
    CREDENTIAL_STATUS_VALIDATE,
    CREDENTIAL_SUBJECT_EXAMPLE,
    CREDENTIAL_SUBJECT_VALIDATE,
    CREDENTIAL_TYPE_EXAMPLE,
    CREDENTIAL_TYPE_VALIDATE,
    RFC3339_DATETIME_EXAMPLE,
    RFC3339_DATETIME_VALIDATE,
    DictOrDictListField,
    DIDKey,
    StrOrDictField,
    Uri,
    UriOrDictField,
)
from ...ld_proofs.constants import (
    CREDENTIALS_CONTEXT_V1_URL,
    CREDENTIALS_CONTEXT_V2_URL,
    VERIFIABLE_CREDENTIAL_TYPE,
)
from .linked_data_proof import LDProof, LinkedDataProofSchema


class VerifiableCredential(BaseModel):
    """Verifiable Credential model."""

    class Meta:
        """VerifiableCredential metadata."""

        schema_class = "CredentialSchema"

    def __init__(
        self,
        context: Optional[List[Union[str, dict]]] = None,
        id: Optional[str] = None,
        type: Optional[List[str]] = None,
        issuer: Optional[Union[dict, str]] = None,
        issuance_date: Optional[str] = None,
        expiration_date: Optional[str] = None,
        valid_from: Optional[str] = None,
        valid_until: Optional[str] = None,
        credential_subject: Optional[Union[dict, List[dict]]] = None,
        credential_status: Optional[Union[dict, List[dict]]] = None,
        proof: Optional[Union[dict, LDProof]] = None,
        **kwargs,
    ) -> None:
        """Initialize the VerifiableCredential instance."""
        self._context = context or [CREDENTIALS_CONTEXT_V1_URL]
        self._id = id
        self._type = type or [VERIFIABLE_CREDENTIAL_TYPE]
        self._issuer = issuer
        self._credential_subject = credential_subject
        self._credential_status = credential_status

        # TODO: proper date parsing
        self._issuance_date = issuance_date
        self._expiration_date = expiration_date
        self._valid_from = valid_from
        self._valid_until = valid_until

        self._proof = proof

        self.extra = kwargs

    @property
    def context(self):
        """Getter for context."""
        return self._context

    @context.setter
    def context(self, context: List[Union[str, dict]]):
        """Setter for context.

        First item must be credentials v1 url
        """
        assert context[0] in [CREDENTIALS_CONTEXT_V1_URL, CREDENTIALS_CONTEXT_V2_URL]

        self._context = context

    def add_context(self, context: Union[str, dict]):
        """Add a context to this credential."""
        self._context.append(context)

    @property
    def context_urls(self) -> List[str]:
        """Getter for context urls."""
        return [context for context in self.context if isinstance(context, str)]

    @property
    def type(self) -> List[str]:
        """Getter for type."""
        return self._type

    @type.setter
    def type(self, type: List[str]):
        """Setter for type.

        First item must be VerifiableCredential
        """
        assert VERIFIABLE_CREDENTIAL_TYPE in type

        self._type = type

    def add_type(self, type: str):
        """Add a type to this credential."""
        self._type.append(type)

    @property
    def id(self):
        """Getter for id."""
        return self._id

    @id.setter
    def id(self, id: Union[str, None]):
        """Setter for id."""
        if id:
            uri_validator = Uri()
            uri_validator(id)

        self._id = id

    @property
    def issuer_id(self) -> Optional[str]:
        """Getter for issuer id."""
        if not self._issuer:
            return None
        elif isinstance(self._issuer, str):
            return self._issuer

        return self._issuer.get("id")

    @issuer_id.setter
    def issuer_id(self, issuer_id: str):
        """Setter for issuer id."""
        uri_validator = Uri()
        uri_validator(issuer_id)

        # Use simple string variant if possible
        if not self._issuer or isinstance(self._issuer, str):
            self._issuer = issuer_id
        else:
            self._issuer["id"] = issuer_id

    @property
    def issuer(self):
        """Getter for issuer."""
        return self._issuer

    @issuer.setter
    def issuer(self, issuer: Union[str, dict]):
        """Setter for issuer."""
        uri_validator = Uri()

        issuer_id = issuer if isinstance(issuer, str) else issuer.get("id")

        if not issuer_id:
            raise ValidationError("Issuer id is required")
        uri_validator(issuer_id)

        self._issuer = issuer

    @property
    def issuance_date(self):
        """Getter for issuance date."""
        return self._issuance_date

    @issuance_date.setter
    def issuance_date(self, date: Union[str, datetime]):
        """Setter for issuance date."""
        if isinstance(date, datetime):
            if not date.tzinfo:
                date = date.replace(tzinfo=tz.UTC)
            date = date.isoformat()

        self._issuance_date = date

    @property
    def expiration_date(self):
        """Getter for expiration date."""
        return self._expiration_date

    @expiration_date.setter
    def expiration_date(self, date: Union[str, datetime, None]):
        """Setter for expiration date."""
        if isinstance(date, datetime):
            if not date.tzinfo:
                date = date.replace(tzinfo=tz.UTC)
            date = date.isoformat()

        self._expiration_date = date

    @property
    def valid_from(self):
        """Getter for valid from date."""
        return self._valid_from

    @valid_from.setter
    def valid_from(self, date: Union[str, datetime]):
        """Setter for valid from date."""
        if isinstance(date, datetime):
            if not date.tzinfo:
                date = date.replace(tzinfo=tz.UTC)
            date = date.isoformat()

        self._valid_from = date

    @property
    def valid_until(self):
        """Getter for valid until date."""
        return self._valid_until

    @valid_until.setter
    def valid_until(self, date: Union[str, datetime, None]):
        """Setter for valid until date."""
        if isinstance(date, datetime):
            if not date.tzinfo:
                date = date.replace(tzinfo=tz.UTC)
            date = date.isoformat()

        self._valid_until = date

    @property
    def credential_subject_ids(self) -> List[str]:
        """Getter for credential subject ids."""
        if not self._credential_subject:
            return []
        elif isinstance(self._credential_subject, dict):
            subject_id = self._credential_subject.get("id")

            return [subject_id] if subject_id else []
        else:
            return [
                subject.get("id")
                for subject in self._credential_subject
                if subject.get("id")
            ]

    @property
    def credential_subject(self):
        """Getter for credential subject."""
        return self._credential_subject

    @credential_subject.setter
    def credential_subject(self, credential_subject: Union[dict, List[dict]]):
        """Setter for credential subject."""

        uri_validator = Uri()

        subjects = (
            [credential_subject]
            if isinstance(credential_subject, dict)
            else credential_subject
        )

        # loop trough all credential subjects and check for valid id uri
        for subject in subjects:
            if subject.get("id"):
                uri_validator(subject.get("id"))

        self._credential_subject = credential_subject

    @property
    def credential_status(self):
        """Getter for credential status."""
        return self._credential_status

    @property
    def proof(self):
        """Getter for proof."""
        return self._proof

    @proof.setter
    def proof(self, proof: LDProof):
        """Setter for proof."""
        self._proof = proof

    def __eq__(self, o: object) -> bool:
        """Check equality."""
        if isinstance(o, VerifiableCredential):
            return (
                self.context == o.context
                and self.id == o.id
                and self.type == o.type
                and self.issuer == o.issuer
                and self.issuance_date == o.issuance_date
                and self.expiration_date == o.expiration_date
                and self.valid_from == o.valid_from
                and self.valid_until == o.valid_until
                and self.credential_subject == o.credential_subject
                and self.credential_status == o.credential_status
                and self.proof == o.proof
                and self.extra == o.extra
            )

        return False


class CredentialSchema(BaseModelSchema):
    """Linked data credential schema.

    Based on https://www.w3.org/TR/vc-data-model

    """

    class Meta:
        """Accept parameter overload."""

        unknown = INCLUDE
        model_class = VerifiableCredential

    context = fields.List(
        UriOrDictField(required=True),
        data_key="@context",
        required=True,
        validate=CREDENTIAL_CONTEXT_VALIDATE,
        metadata={
            "description": "The JSON-LD context of the credential",
            "example": CREDENTIAL_CONTEXT_EXAMPLE,
        },
    )

    id = fields.Str(
        required=False,
        validate=Uri(),
        metadata={
            "description": "The ID of the credential",
            "example": "http://example.edu/credentials/1872",
        },
    )

    type = fields.List(
        fields.Str(required=True),
        required=True,
        validate=CREDENTIAL_TYPE_VALIDATE,
        metadata={
            "description": "The JSON-LD type of the credential",
            "example": CREDENTIAL_TYPE_EXAMPLE,
        },
    )

    issuer = StrOrDictField(
        required=True,
        metadata={
            "description": (
                "The JSON-LD Verifiable Credential Issuer. Either string of object with"
                " id field."
            ),
            "example": DIDKey.EXAMPLE,
        },
    )

    issuance_date = fields.Str(
        data_key="issuanceDate",
        required=False,
        validate=RFC3339_DATETIME_VALIDATE,
        metadata={
            "description": "The issuance date",
            "example": RFC3339_DATETIME_EXAMPLE,
        },
    )

    expiration_date = fields.Str(
        data_key="expirationDate",
        required=False,
        validate=RFC3339_DATETIME_VALIDATE,
        metadata={
            "description": "The expiration date",
            "example": RFC3339_DATETIME_EXAMPLE,
        },
    )

    valid_from = fields.Str(
        data_key="validFrom",
        required=False,
        validate=RFC3339_DATETIME_VALIDATE,
        metadata={
            "description": "The valid from date",
            "example": RFC3339_DATETIME_EXAMPLE,
        },
    )

    valid_until = fields.Str(
        data_key="validUntil",
        required=False,
        validate=RFC3339_DATETIME_VALIDATE,
        metadata={
            "description": "The valid until date",
            "example": RFC3339_DATETIME_EXAMPLE,
        },
    )

    credential_subject = DictOrDictListField(
        required=True,
        data_key="credentialSubject",
        validate=CREDENTIAL_SUBJECT_VALIDATE,
        metadata={"example": CREDENTIAL_SUBJECT_EXAMPLE},
    )

    credential_status = DictOrDictListField(
        required=False,
        data_key="credentialStatus",
        validate=CREDENTIAL_STATUS_VALIDATE,
        metadata={"example": CREDENTIAL_STATUS_EXAMPLE},
    )

    proof = fields.Nested(
        LinkedDataProofSchema(),
        required=False,
        metadata={
            "description": "The proof of the credential",
            "example": {
                "type": "Ed25519Signature2018",
                "verificationMethod": (
                    "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38Ee"
                    "fXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                ),
                "created": "2019-12-11T03:50:55",
                "proofPurpose": "assertionMethod",
                "jws": (
                    "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0JiNjQiXX0..lKJU0Df_k"
                    "eblRKhZAS9Qq6zybm-HqUXNVZ8vgEPNTAjQKBhQDxvXNo7nvtUBb_Eq1Ch6YBKY5qBQ"
                ),
            },
        },
    )

    @post_dump(pass_original=True)
    def add_unknown_properties(self, data: dict, original, **kwargs):
        """Add back unknown properties before outputting."""

        data.update(original.extra)

        return data


class VerifiableCredentialSchema(CredentialSchema):
    """Linked data verifiable credential schema.

    Based on https://www.w3.org/TR/vc-data-model

    """

    proof = fields.Nested(
        LinkedDataProofSchema(),
        required=True,
        metadata={
            "description": "The proof of the credential",
            "example": {
                "type": "Ed25519Signature2018",
                "verificationMethod": (
                    "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38Ee"
                    "fXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                ),
                "created": "2019-12-11T03:50:55",
                "proofPurpose": "assertionMethod",
                "jws": (
                    "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0JiNjQiXX0..lKJU0Df_k"
                    "eblRKhZAS9Qq6zybm-HqUXNVZ8vgEPNTAjQKBhQDxvXNo7nvtUBb_Eq1Ch6YBKY5qBQ"
                ),
            },
        },
    )



"""Utilities for creating out-of-band messages."""

import json
from urllib.parse import quote, urljoin

from ..messaging.agent_message import AgentMessage
from ..wallet.did_info import DIDInfo
from ..wallet.util import str_to_b64


def serialize_outofband(message: AgentMessage, did: DIDInfo, endpoint: str) -> str:
    """Serialize the agent message as an out-of-band message.

    Returns:
        An OOB message in URL format.

    """
    body = message.serialize()
    # FIXME no support for routing keys
    body["~services"] = {
        "recipientKeys": [did.verkey],
        "routingKeys": [],
        "serviceEndpoint": endpoint,
    }
    d_m = quote(str_to_b64(json.dumps(body)))
    return urljoin(endpoint, "?d_m={}".format(d_m))



"""Test LegacyIndyRegistry."""

import json
import re
from unittest import IsolatedAsyncioTestCase

import pytest
from anoncreds import (
    CredentialDefinition,
    RevocationRegistryDefinition,
    RevocationRegistryDefinitionPrivate,
    Schema,
)
from base58 import alphabet

from .....anoncreds.base import AnonCredsSchemaAlreadyExists
from .....anoncreds.default.legacy_indy import registry as test_module
from .....anoncreds.issuer import AnonCredsIssuer
from .....askar.profile_anon import AskarAnonCredsProfileSession
from .....connections.models.conn_record import ConnRecord
from .....core.event_bus import EventBus
from .....ledger.base import BaseLedger
from .....ledger.error import LedgerObjectAlreadyExistsError
from .....ledger.multiple_ledger.ledger_requests_executor import (
    IndyLedgerRequestsExecutor,
)
from .....messaging.responder import BaseResponder
from .....protocols.endorse_transaction.v1_0.manager import TransactionManager
from .....protocols.endorse_transaction.v1_0.models.transaction_record import (
    TransactionRecord,
)
from .....revocation_anoncreds.models.issuer_cred_rev_record import IssuerCredRevRecord
from .....tests import mock
from .....utils.testing import create_test_profile
from ....models.credential_definition import (
    CredDef,
    CredDefResult,
    CredDefValue,
    CredDefValuePrimary,
)
from ....models.revocation import (
    RevList,
    RevListResult,
    RevRegDef,
    RevRegDefResult,
    RevRegDefState,
    RevRegDefValue,
)
from ....models.schema import AnonCredsSchema, GetSchemaResult, SchemaResult

B58 = alphabet if isinstance(alphabet, str) else alphabet.decode("ascii")
INDY_DID = rf"^(did:sov:)?[{B58}]{{21,22}}$"
INDY_SCHEMA_ID = rf"^[{B58}]{{21,22}}:2:.+:[0-9.]+$"
INDY_SCHEMA_TXN_ID = r"^[0-9.]+$"
INDY_CRED_DEF_ID = (
    rf"^([{B58}]{{21,22}})"  # issuer DID
    f":3"  # cred def id marker
    f":CL"  # sig alg
    rf":(([1-9][0-9]*)|([{B58}]{{21,22}}:2:.+:[0-9.]+))"  # schema txn / id
    f":(.+)?$"  # tag
)
INDY_REV_REG_DEF_ID = (
    rf"^([{B58}]{{21,22}}):4:"
    rf"([{B58}]{{21,22}}):3:"
    rf"CL:(([1-9][0-9]*)|([{B58}]{{21,22}}:2:.+:[0-9.]+))(:.+)?:"
    rf"CL_ACCUM:(.+$)"
)
SUPPORTED_ID_REGEX = re.compile(
    rf"{INDY_DID}|{INDY_SCHEMA_ID}|{INDY_SCHEMA_TXN_ID}|{INDY_CRED_DEF_ID}|{INDY_REV_REG_DEF_ID}"
)

TEST_INDY_DID = "WgWxqztrNooG92RXvxSTWv"
TEST_INDY_DID_1 = "did:sov:WgWxqztrNooG92RXvxSTWv"
TEST_INDY_SCHEMA_ID = "WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0"
TEST_INDY_CRED_DEF_ID = "WgWxqztrNooG92RXvxSTWv:3:CL:20:tag"
TEST_INDY_REV_REG_DEF_ID = (
    "WgWxqztrNooG92RXvxSTWv:4:WgWxqztrNooG92RXvxSTWv:3:CL:20:tag:CL_ACCUM:0"
)

mock_schema = AnonCredsSchema(
    issuer_id="did:indy:sovrin:SGrjRL82Y9ZZbzhUDXokvQ",
    attr_names=["name", "age", "vmax"],
    name="test_schema",
    version="1.0",
)


class MockTxn:
    def to_json(self):
        return json.dumps(self.__dict__)


class MockRevRegDefEntry:
    def __init__(self, name="name"):
        self.name = name

    tags = {
        "state": RevRegDefState.STATE_ACTION,
    }
    value = "mock_value"
    value_json = {
        "value": {
            "maxCredNum": 100,
            "publicKeys": {"accumKey": {"z": "1 0BB...386"}},
            "tailsHash": "string",
            "tailsLocation": "string",
        },
        "credDefId": "CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
        "issuerId": "CsQY9MGeD3CQP4EyuVFo5m",
        "revocDefType": "CL_ACCUM",
        "tag": "string",
    }


class MockCredDefEntry:
    value_json = {}


class MockRevListEntry:
    tags = {}
    value = "mock_value"
    value_json = {
        "issuerId": "CsQY9MGeD3CQP4EyuVFo5m",
        "revRegDefId": "4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
        "revocationList": [0, 1, 0, 0],
        "currentAccumulator": "21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
        "timestamp": 1669640864487,
    }

    def to_json(self):
        return self.value_json

    def to_dict(self):
        return self.value_json


@pytest.mark.anoncreds
class TestLegacyIndyRegistry(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile(
            settings={"wallet.type": "askar-anoncreds"},
        )
        self.registry = test_module.LegacyIndyRegistry()

    async def test_supported_did_regex(self):
        """Test the supported_did_regex."""

        assert self.registry.supported_identifiers_regex == SUPPORTED_ID_REGEX
        assert bool(self.registry.supported_identifiers_regex.match(TEST_INDY_DID))
        assert bool(self.registry.supported_identifiers_regex.match(TEST_INDY_DID_1))
        assert bool(self.registry.supported_identifiers_regex.match(TEST_INDY_SCHEMA_ID))
        assert bool(
            self.registry.supported_identifiers_regex.match(TEST_INDY_REV_REG_DEF_ID)
        )

    async def test_register_schema_no_endorsement(self):
        self.profile.inject_or = mock.MagicMock(
            return_value=mock.CoroutineMock(
                send_schema_anoncreds=mock.CoroutineMock(return_value=1)
            )
        )

        result = await self.registry.register_schema(self.profile, mock_schema, {})

        assert isinstance(result, SchemaResult)

    @mock.patch.object(
        ConnRecord,
        "retrieve_by_id",
        return_value=mock.CoroutineMock(
            metadata_get=mock.CoroutineMock(return_value={"endorser_did": "test_did"})
        ),
    )
    async def test_register_schema_with_author_role(self, mock_endorser_conn_record):
        self.profile.inject_or = mock.MagicMock(
            return_value=mock.CoroutineMock(
                send_schema_anoncreds=mock.CoroutineMock(
                    return_value=(
                        "test_schema_id",
                        {
                            "signed_txn": "test_signed_txn",
                        },
                    )
                )
            )
        )
        self.profile.settings.set_value("endorser.author", True)

        result = await self.registry.register_schema(
            self.profile, mock_schema, {"endorser_connection_id": "test_connection_id"}
        )

        assert result.registration_metadata["txn"] is not None
        assert mock_endorser_conn_record.called

    @mock.patch.object(
        ConnRecord,
        "retrieve_by_id",
        return_value=mock.CoroutineMock(
            metadata_get=mock.CoroutineMock(return_value={"endorser_did": "test_did"})
        ),
    )
    async def test_register_schema_already_exists(self, mock_endorser_conn_record):
        self.profile.inject_or = mock.MagicMock(
            return_value=mock.CoroutineMock(
                send_schema_anoncreds=mock.CoroutineMock(
                    side_effect=LedgerObjectAlreadyExistsError(
                        "test", "test", mock_schema
                    )
                )
            )
        )
        self.profile.settings.set_value("endorser.author", True)

        with self.assertRaises(AnonCredsSchemaAlreadyExists):
            await self.registry.register_schema(
                self.profile,
                mock_schema,
                {"endorser_connection_id": "test_connection_id"},
            )

    @mock.patch.object(
        ConnRecord,
        "retrieve_by_id",
        return_value=mock.CoroutineMock(
            metadata_get=mock.CoroutineMock(return_value={"endorser_did": "test_did"})
        ),
    )
    async def test_register_schema_with_create_trasaction_param(
        self, mock_endorser_conn_record
    ):
        self.profile.inject_or = mock.MagicMock(
            return_value=mock.CoroutineMock(
                send_schema_anoncreds=mock.CoroutineMock(
                    return_value=(
                        "test_schema_id",
                        {
                            "signed_txn": "test_signed_txn",
                        },
                    )
                )
            )
        )

        result = await self.registry.register_schema(
            self.profile,
            mock_schema,
            {
                "endorser_connection_id": "test_connection_id",
                "create_transaction_for_endorser": True,
            },
        )

        assert result.registration_metadata["txn"] is not None
        assert mock_endorser_conn_record.called

    @mock.patch.object(
        ConnRecord,
        "retrieve_by_id",
        return_value=mock.CoroutineMock(
            metadata_get=mock.CoroutineMock(return_value={"endorser_did": "test_did"})
        ),
    )
    @mock.patch.object(
        TransactionManager,
        "create_request",
    )
    async def test_register_schema_with_author_role_and_create_request(
        self, mock_create_request, mock_endorser_conn_record
    ):
        class MockTransaction:
            def __init__(self, txn):
                self.txn = txn

            def serialize(self):
                return json.dumps(self.txn)

        self.profile.inject_or = mock.MagicMock(
            return_value=mock.CoroutineMock(
                send_schema_anoncreds=mock.CoroutineMock(
                    return_value=(
                        "test_schema_id",
                        {
                            "signed_txn": "test_signed_txn",
                        },
                    )
                )
            )
        )

        mock_create_request.return_value = (
            MockTransaction({"test": "test"}),
            "transaction_request",
        )

        self.profile.settings.set_value("endorser.author", True)
        self.profile.settings.set_value("endorser.auto_request", True)
        self.profile.context.injector.bind_instance(
            BaseResponder, mock.MagicMock(BaseResponder, autospec=True)
        )

        result = await self.registry.register_schema(
            self.profile, mock_schema, {"endorser_connection_id": "test_connection_id"}
        )

        assert result.registration_metadata["txn"] is not None
        assert mock_create_request.called
        assert self.profile.context.injector.get_provider(
            BaseResponder
        )._instance.send.called

    @mock.patch.object(
        AnonCredsIssuer, "credential_definition_in_wallet", return_value=False
    )
    async def test_register_credential_definition_no_endorsement(self, mock_in_wallet):
        schema = Schema.create(
            name="MYCO Biomarker",
            version="1.0",
            issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
            attr_names=["attr1", "attr2"],
        )

        schema_result = GetSchemaResult(
            schema_id="schema-id",
            schema=schema,
            schema_metadata={
                "seqNo": 1,
            },
            resolution_metadata={},
        )

        cred_def = CredDef(
            issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
            schema_id="CsQY9MGeD3CQP4EyuVFo5m:2:MYCO Biomarker:0.0.3",
            tag="default",
            type="CL",
            value=CredDefValue(primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z")),
        )

        mock_base_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_base_ledger.send_credential_definition_anoncreds = mock.CoroutineMock(
            return_value=2
        )
        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock_base_ledger,
        )

        result = await self.registry.register_credential_definition(
            self.profile, schema_result, cred_def, {}
        )

        assert isinstance(result, CredDefResult)
        assert mock_in_wallet.called

    @mock.patch.object(
        AnonCredsIssuer, "credential_definition_in_wallet", return_value=False
    )
    @mock.patch.object(
        ConnRecord,
        "retrieve_by_id",
        return_value=mock.CoroutineMock(
            metadata_get=mock.CoroutineMock(return_value={"endorser_did": "test_did"})
        ),
    )
    @mock.patch.object(
        TransactionManager,
        "create_record",
        return_value=TransactionRecord(),
    )
    async def test_register_credential_definition_with_author_role(
        self, mock_create_record, mock_get_endorser_info, mock_in_wallet
    ):
        schema = Schema.create(
            name="MYCO Biomarker",
            version="1.0",
            issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
            attr_names=["attr1", "attr2"],
        )

        schema_result = GetSchemaResult(
            schema_id="schema-id",
            schema=schema,
            schema_metadata={
                "seqNo": 1,
            },
            resolution_metadata={},
        )

        cred_def = CredDef(
            issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
            schema_id="CsQY9MGeD3CQP4EyuVFo5m:2:MYCO Biomarker:0.0.3",
            tag="default",
            type="CL",
            value=CredDefValue(primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z")),
        )

        mock_base_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_base_ledger.send_credential_definition_anoncreds = mock.CoroutineMock(
            return_value=("id", {"signed_txn": "txn"})
        )
        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock_base_ledger,
        )
        self.profile.settings.set_value("endorser.author", True)

        result = await self.registry.register_credential_definition(
            self.profile,
            schema_result,
            cred_def,
            {"endorser_connection_id": "test_connection_id"},
        )

        assert isinstance(result, CredDefResult)
        assert result.job_id is not None
        assert mock_in_wallet.called
        assert mock_get_endorser_info.called
        assert mock_create_record.called

    @mock.patch.object(
        AnonCredsIssuer, "credential_definition_in_wallet", return_value=False
    )
    @mock.patch.object(
        ConnRecord,
        "retrieve_by_id",
        return_value=mock.CoroutineMock(
            metadata_get=mock.CoroutineMock(return_value={"endorser_did": "test_did"})
        ),
    )
    @mock.patch.object(
        TransactionManager,
        "create_record",
        return_value=TransactionRecord(),
    )
    async def test_register_credential_definition_with_create_transaction_option(
        self, mock_create_record, mock_get_endorser_info, mock_in_wallet
    ):
        schema = Schema.create(
            name="MYCO Biomarker",
            version="1.0",
            issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
            attr_names=["attr1", "attr2"],
        )

        schema_result = GetSchemaResult(
            schema_id="schema-id",
            schema=schema,
            schema_metadata={
                "seqNo": 1,
            },
            resolution_metadata={},
        )

        cred_def = CredDef(
            issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
            schema_id="CsQY9MGeD3CQP4EyuVFo5m:2:MYCO Biomarker:0.0.3",
            tag="default",
            type="CL",
            value=CredDefValue(primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z")),
        )

        mock_base_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_base_ledger.send_credential_definition_anoncreds = mock.CoroutineMock(
            return_value=("id", {"signed_txn": "txn"})
        )
        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock_base_ledger,
        )

        result = await self.registry.register_credential_definition(
            self.profile,
            schema_result,
            cred_def,
            {
                "endorser_connection_id": "test_connection_id",
                "create_transaction_for_endorser": True,
            },
        )

        assert isinstance(result, CredDefResult)
        assert result.job_id is not None
        assert mock_in_wallet.called
        assert mock_get_endorser_info.called
        assert mock_create_record.called

    @mock.patch.object(
        AnonCredsIssuer, "credential_definition_in_wallet", return_value=False
    )
    @mock.patch.object(
        ConnRecord,
        "retrieve_by_id",
        return_value=mock.CoroutineMock(
            metadata_get=mock.CoroutineMock(return_value={"endorser_did": "test_did"})
        ),
    )
    @mock.patch.object(
        TransactionManager,
        "create_record",
        return_value=TransactionRecord(),
    )
    @mock.patch.object(
        TransactionManager,
        "create_request",
        return_value=(TransactionRecord(), "transaction_request"),
    )
    async def test_register_credential_definition_with_transaction_and_auto_request(
        self,
        mock_create_request,
        mock_create_record,
        mock_get_endorser_info,
        mock_in_wallet,
    ):
        schema = Schema.create(
            name="MYCO Biomarker",
            version="1.0",
            issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
            attr_names=["attr1", "attr2"],
        )

        schema_result = GetSchemaResult(
            schema_id="schema-id",
            schema=schema,
            schema_metadata={
                "seqNo": 1,
            },
            resolution_metadata={},
        )

        cred_def = CredDef(
            issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
            schema_id="CsQY9MGeD3CQP4EyuVFo5m:2:MYCO Biomarker:0.0.3",
            tag="default",
            type="CL",
            value=CredDefValue(primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z")),
        )

        mock_base_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_base_ledger.send_credential_definition_anoncreds = mock.CoroutineMock(
            return_value=("id", {"signed_txn": "txn"})
        )
        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock_base_ledger,
        )

        self.profile.context.injector.bind_instance(
            BaseResponder,
            mock.MagicMock(BaseResponder, autospec=True),
        )
        self.profile.settings.set_value("endorser.auto_request", True)

        result = await self.registry.register_credential_definition(
            self.profile,
            schema_result,
            cred_def,
            {
                "endorser_connection_id": "test_connection_id",
                "create_transaction_for_endorser": True,
            },
        )

        assert isinstance(result, CredDefResult)
        assert result.job_id is not None
        assert mock_in_wallet.called
        assert mock_get_endorser_info.called
        assert mock_create_record.called
        assert mock_create_request.called
        assert self.profile.context.injector.get_provider(
            BaseResponder
        )._instance.send.called

    async def test_register_revocation_registry_definition_no_endorsement(self):
        mock_base_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_base_ledger.send_revoc_reg_def = mock.CoroutineMock(return_value=1)
        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock_base_ledger,
        )
        result = await self.registry.register_revocation_registry_definition(
            self.profile,
            RevRegDef(
                tag="tag",
                cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                value=RevRegDefValue(
                    max_cred_num=100,
                    public_keys={
                        "accum_key": {"z": "1 0BB...386"},
                    },
                    tails_hash="not-correct-hash",
                    tails_location="http://tails-server.com",
                ),
                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                type="CL_ACCUM",
            ),
            {},
        )

        assert isinstance(result, RevRegDefResult)
        assert self.profile.context.injector.get_provider(
            BaseLedger
        )._instance.send_revoc_reg_def.called

    @mock.patch.object(
        ConnRecord,
        "retrieve_by_id",
        return_value=mock.CoroutineMock(
            metadata_get=mock.CoroutineMock(return_value={"endorser_did": "test_did"})
        ),
    )
    @mock.patch.object(
        TransactionManager,
        "create_record",
        return_value=TransactionRecord(),
    )
    @mock.patch.object(
        TransactionManager,
        "create_request",
        return_value=(TransactionRecord(), "transaction_request"),
    )
    async def test_register_revocation_registry_definition_with_author_role(
        self, mock_create_request, mock_create_record, mock_endorser_connection
    ):
        self.profile.settings.set_value("endorser.author", True)
        mock_base_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_base_ledger.send_revoc_reg_def = mock.CoroutineMock(
            return_value=("id", {"signed_txn": "txn"})
        )
        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock_base_ledger,
        )
        self.profile.context.injector.bind_instance(
            BaseResponder,
            mock.MagicMock(BaseResponder, autospec=True),
        )

        self.profile.settings.set_value("endorser.auto_request", True)

        result = await self.registry.register_revocation_registry_definition(
            self.profile,
            RevRegDef(
                tag="tag",
                cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                value=RevRegDefValue(
                    max_cred_num=100,
                    public_keys={
                        "accum_key": {"z": "1 0BB...386"},
                    },
                    tails_hash="not-correct-hash",
                    tails_location="http://tails-server.com",
                ),
                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                type="CL_ACCUM",
            ),
            {
                "endorser_connection_id": "test_connection_id",
            },
        )

        assert isinstance(result, RevRegDefResult)
        assert self.profile.context.injector.get_provider(
            BaseLedger
        )._instance.send_revoc_reg_def.called
        assert mock_create_record.called
        assert mock_create_request.called
        assert self.profile.context.injector.get_provider(
            BaseResponder
        )._instance.send.called

    @mock.patch.object(
        ConnRecord,
        "retrieve_by_id",
        return_value=mock.CoroutineMock(
            metadata_get=mock.CoroutineMock(return_value={"endorser_did": "test_did"})
        ),
    )
    @mock.patch.object(
        TransactionManager,
        "create_record",
        return_value=TransactionRecord(),
    )
    async def test_register_revocation_registry_definition_with_create_transaction_option(
        self, mock_create_record, mock_endorser_connection
    ):
        mock_base_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_base_ledger.send_revoc_reg_def = mock.CoroutineMock(
            return_value=("id", {"signed_txn": "txn"})
        )
        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock_base_ledger,
        )

        result = await self.registry.register_revocation_registry_definition(
            self.profile,
            RevRegDef(
                tag="tag",
                cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                value=RevRegDefValue(
                    max_cred_num=100,
                    public_keys={
                        "accum_key": {"z": "1 0BB...386"},
                    },
                    tails_hash="not-correct-hash",
                    tails_location="http://tails-server.com",
                ),
                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                type="CL_ACCUM",
            ),
            {
                "endorser_connection_id": "test_connection_id",
                "create_transaction_for_endorser": True,
            },
        )

        assert isinstance(result, RevRegDefResult)
        assert self.profile.context.injector.get_provider(
            BaseLedger
        )._instance.send_revoc_reg_def.called
        assert mock_create_record.called

    @mock.patch.object(
        ConnRecord,
        "retrieve_by_id",
        return_value=mock.CoroutineMock(
            metadata_get=mock.CoroutineMock(return_value={"endorser_did": "test_did"})
        ),
    )
    @mock.patch.object(
        TransactionManager,
        "create_record",
        return_value=TransactionRecord(),
    )
    async def test_register_revocation_registry_definition_with_create_transaction_and_auto_request(
        self, mock_create_record, mock_endorser_connection
    ):
        mock_base_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_base_ledger.send_revoc_reg_def = mock.CoroutineMock(
            return_value=("id", {"signed_txn": "txn"})
        )
        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock_base_ledger,
        )

        result = await self.registry.register_revocation_registry_definition(
            self.profile,
            RevRegDef(
                tag="tag",
                cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                value=RevRegDefValue(
                    max_cred_num=100,
                    public_keys={
                        "accum_key": {"z": "1 0BB...386"},
                    },
                    tails_hash="not-correct-hash",
                    tails_location="http://tails-server.com",
                ),
                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                type="CL_ACCUM",
            ),
            {
                "endorser_connection_id": "test_connection_id",
                "create_transaction_for_endorser": True,
            },
        )

        assert isinstance(result, RevRegDefResult)
        assert self.profile.context.injector.get_provider(
            BaseLedger
        )._instance.send_revoc_reg_def.called
        assert mock_create_record.called

    async def test_txn_submit(self):
        mock_base_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_base_ledger.txn_submit = mock.CoroutineMock(return_value="transaction_id")
        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock_base_ledger,
        )

        async with self.profile.session() as session:
            ledger = session.inject(BaseLedger)
            result = await self.registry.txn_submit(ledger, "test_txn")
            assert result == "transaction_id"

    @mock.patch.object(
        IndyLedgerRequestsExecutor,
        "get_ledger_for_identifier",
        return_value=(
            "id",
            mock.MagicMock(
                send_revoc_reg_entry=mock.CoroutineMock(return_value="transaction_id")
            ),
        ),
    )
    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_register_revocation_list_no_endorsement(
        self, mock_handle, mock_send_revoc_reg_entry
    ):
        self.profile.inject_or = mock.MagicMock()
        mock_handle.fetch = mock.CoroutineMock(
            side_effect=[
                mock.CoroutineMock(return_value=None),
                mock.CoroutineMock(return_value=None),
                mock.CoroutineMock(return_value=None),
            ]
        )

        result = await self.registry.register_revocation_list(
            self.profile,
            RevRegDef(
                tag="tag",
                cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                value=RevRegDefValue(
                    max_cred_num=100,
                    public_keys={
                        "accum_key": {"z": "1 0BB...386"},
                    },
                    tails_hash="not-correct-hash",
                    tails_location="http://tails-server.com",
                ),
                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                type="CL_ACCUM",
            ),
            RevList(
                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                current_accumulator="21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
                revocation_list=[0, 1, 1, 0],
                timestamp=1669640864487,
                rev_reg_def_id="4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
            ),
            {},
        )

        assert isinstance(result, RevListResult)
        assert mock_send_revoc_reg_entry.called

    @mock.patch.object(
        ConnRecord,
        "retrieve_by_id",
        return_value=mock.CoroutineMock(
            metadata_get=mock.CoroutineMock(return_value={"endorser_did": "test_did"})
        ),
    )
    @mock.patch.object(
        TransactionManager,
        "create_record",
        return_value=TransactionRecord(),
    )
    @mock.patch.object(
        IndyLedgerRequestsExecutor,
        "get_ledger_for_identifier",
        return_value=(
            "id",
            mock.MagicMock(
                send_revoc_reg_entry=mock.CoroutineMock(
                    return_value=(
                        "rev_reg_def_id",
                        {
                            "signed_txn": "txn",
                        },
                    )
                )
            ),
        ),
    )
    async def test_register_revocation_list_with_author_role(
        self, mock_send_revoc_reg_entry, mock_create_record, _
    ):
        self.profile.inject_or = mock.MagicMock()
        self.profile.settings.set_value("endorser.author", True)

        result = await self.registry.register_revocation_list(
            self.profile,
            RevRegDef(
                tag="tag",
                cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                value=RevRegDefValue(
                    max_cred_num=100,
                    public_keys={
                        "accum_key": {"z": "1 0BB...386"},
                    },
                    tails_hash="not-correct-hash",
                    tails_location="http://tails-server.com",
                ),
                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                type="CL_ACCUM",
            ),
            RevList(
                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                current_accumulator="21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
                revocation_list=[0, 1, 1, 0],
                timestamp=1669640864487,
                rev_reg_def_id="4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
            ),
            {
                "endorser_connection_id": "test_connection_id",
            },
        )

        assert isinstance(result, RevListResult)
        assert mock_create_record.called
        assert mock_send_revoc_reg_entry.called

    @mock.patch.object(
        ConnRecord,
        "retrieve_by_id",
        return_value=mock.CoroutineMock(
            metadata_get=mock.CoroutineMock(return_value={"endorser_did": "test_did"})
        ),
    )
    @mock.patch.object(
        TransactionManager,
        "create_record",
        return_value=TransactionRecord(),
    )
    @mock.patch.object(
        IndyLedgerRequestsExecutor,
        "get_ledger_for_identifier",
        return_value=(
            "id",
            mock.MagicMock(
                send_revoc_reg_entry=mock.CoroutineMock(
                    return_value=(
                        "rev_reg_def_id",
                        {
                            "signed_txn": "txn",
                        },
                    )
                )
            ),
        ),
    )
    async def test_register_revocation_list_with_create_transaction_option(
        self, mock_send_revoc_reg_entry, mock_create_record, _
    ):
        self.profile.inject_or = mock.MagicMock()

        result = await self.registry.register_revocation_list(
            self.profile,
            RevRegDef(
                tag="tag",
                cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                value=RevRegDefValue(
                    max_cred_num=100,
                    public_keys={
                        "accum_key": {"z": "1 0BB...386"},
                    },
                    tails_hash="not-correct-hash",
                    tails_location="http://tails-server.com",
                ),
                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                type="CL_ACCUM",
            ),
            RevList(
                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                current_accumulator="21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
                revocation_list=[0, 1, 1, 0],
                timestamp=1669640864487,
                rev_reg_def_id="4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
            ),
            {
                "endorser_connection_id": "test_connection_id",
                "create_transaction_for_endorser": True,
            },
        )

        assert isinstance(result, RevListResult)
        assert mock_create_record.called
        assert mock_send_revoc_reg_entry.called

    @mock.patch.object(
        ConnRecord,
        "retrieve_by_id",
        return_value=mock.CoroutineMock(
            metadata_get=mock.CoroutineMock(return_value={"endorser_did": "test_did"})
        ),
    )
    @mock.patch.object(
        TransactionManager,
        "create_record",
        return_value=TransactionRecord(),
    )
    @mock.patch.object(
        TransactionManager,
        "create_request",
        return_value=(TransactionRecord(), "transaction_request"),
    )
    @mock.patch.object(
        IndyLedgerRequestsExecutor,
        "get_ledger_for_identifier",
        return_value=(
            "id",
            mock.MagicMock(
                send_revoc_reg_entry=mock.CoroutineMock(
                    return_value=(
                        "rev_reg_def_id",
                        {
                            "signed_txn": "txn",
                        },
                    )
                )
            ),
        ),
    )
    async def test_register_revocation_list_with_create_transaction_option_and_auto_request(
        self, mock_send_revoc_reg_entry, mock_create_request, mock_create_record, _
    ):
        self.profile.inject_or = mock.MagicMock()
        self.profile.context.injector.bind_instance(
            BaseResponder,
            mock.MagicMock(BaseResponder, autospec=True),
        )
        self.profile.settings.set_value("endorser.auto_request", True)

        result = await self.registry.register_revocation_list(
            self.profile,
            RevRegDef(
                tag="tag",
                cred_def_id="CsQY9MGeD3CQP4EyuVFo5m:3:CL:14951:MYCO_Biomarker",
                value=RevRegDefValue(
                    max_cred_num=100,
                    public_keys={
                        "accum_key": {"z": "1 0BB...386"},
                    },
                    tails_hash="not-correct-hash",
                    tails_location="http://tails-server.com",
                ),
                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                type="CL_ACCUM",
            ),
            RevList(
                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                current_accumulator="21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
                revocation_list=[0, 1, 1, 0],
                timestamp=1669640864487,
                rev_reg_def_id="4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
            ),
            {
                "endorser_connection_id": "test_connection_id",
                "create_transaction_for_endorser": True,
            },
        )

        assert isinstance(result, RevListResult)
        assert mock_create_record.called
        assert mock_create_request.called
        assert mock_send_revoc_reg_entry.called
        assert self.profile.context.injector.get_provider(
            BaseResponder
        )._instance.send.called

    @mock.patch.object(
        IndyLedgerRequestsExecutor,
        "get_ledger_for_identifier",
        return_value=(
            "id",
            mock.MagicMock(
                get_revoc_reg_delta=mock.CoroutineMock(
                    return_value=(
                        {
                            "value": {
                                "accum": "21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
                            }
                        },
                        123,
                    )
                )
            ),
        ),
    )
    @mock.patch.object(
        test_module.LegacyIndyRegistry,
        "_sync_wallet_rev_list_with_issuer_cred_rev_records",
        mock.CoroutineMock(
            return_value=RevList(
                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                current_accumulator="2 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
                revocation_list=[1, 0, 1, 0],
                timestamp=1669640864487,
                rev_reg_def_id="4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
            )
        ),
    )
    @mock.patch(
        "acapy_agent.anoncreds.default.legacy_indy.registry.generate_ledger_rrrecovery_txn",
        mock.CoroutineMock(return_value=MockTxn()),
    )
    async def test_fix_ledger_entry(self, *_):
        mock_base_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_base_ledger.send_revoc_reg_entry = mock.CoroutineMock(return_value={})
        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock_base_ledger,
        )

        self.profile.context.injector.bind_instance(
            EventBus,
            {},
        )

        async with self.profile.transaction() as txn:
            issuer_cr_rec = IssuerCredRevRecord(
                state=IssuerCredRevRecord.STATE_ISSUED,
                cred_ex_id="cred_ex_id",
                cred_ex_version=IssuerCredRevRecord.VERSION_1,
                rev_reg_id="4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
                cred_rev_id="cred_rev_id",
            )
            await issuer_cr_rec.save(
                txn,
                reason=("Testing"),
            )

        self.profile.inject_or = mock.MagicMock()
        result = await self.registry.fix_ledger_entry(
            self.profile,
            RevList(
                issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                current_accumulator="21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
                revocation_list=[0, 1, 1, 0],
                timestamp=1669640864487,
                rev_reg_def_id="4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
            ),
            True,
            '{"reqSignature":{},"txn":{"data":{"data":{"alias":"Node1","blskey":"4N8aUNHSgjQVgkpm8nhNEfDf6txHznoYREg9kirmJrkivgL4oSEimFF6nsQ6M41QvhM2Z33nves5vfSn9n1UwNFJBYtWVnHYMATn76vLuL3zU88KyeAYcHfsih3He6UHcXDxcaecHVz6jhCYz1P2UZn2bDVruL5wXpehgBfBaLKm3Ba","blskey_pop":"RahHYiCvoNCtPTrVtP7nMC5eTYrsUA8WjXbdhNc8debh1agE9bGiJxWBXYNFbnJXoXhWFMvyqhqhRoq737YQemH5ik9oL7R4NTTCz2LEZhkgLJzB3QRQqJyBNyv7acbdHrAT8nQ9UkLbaVL9NBpnWXBTw4LEMePaSHEw66RzPNdAX1","client_ip":"172.17.0.2","client_port":9702,"node_ip":"172.17.0.2","node_port":9701,"services":["VALIDATOR"]},"dest":"Gw6pDLhcBcoQesN72qfotTgFa7cbuqZpkX3Xo6pLhPhv"},"metadata":{"from":"Th7MpTaRZVRYnPiabds81Y"},"type":"0"},"txnMetadata":{"seqNo":1,"txnId":"fea82e10e894419fe2bea7d96296a6d46f50f93f9eeda954ec461b2ed2950b62"},"ver":"1"}\n{"reqSignature":{},"txn":{"data":{"data":{"alias":"Node2","blskey":"37rAPpXVoxzKhz7d9gkUe52XuXryuLXoM6P6LbWDB7LSbG62Lsb33sfG7zqS8TK1MXwuCHj1FKNzVpsnafmqLG1vXN88rt38mNFs9TENzm4QHdBzsvCuoBnPH7rpYYDo9DZNJePaDvRvqJKByCabubJz3XXKbEeshzpz4Ma5QYpJqjk","blskey_pop":"Qr658mWZ2YC8JXGXwMDQTzuZCWF7NK9EwxphGmcBvCh6ybUuLxbG65nsX4JvD4SPNtkJ2w9ug1yLTj6fgmuDg41TgECXjLCij3RMsV8CwewBVgVN67wsA45DFWvqvLtu4rjNnE9JbdFTc1Z4WCPA3Xan44K1HoHAq9EVeaRYs8zoF5","client_ip":"172.17.0.2","client_port":9704,"node_ip":"172.17.0.2","node_port":9703,"services":["VALIDATOR"]},"dest":"8ECVSk179mjsjKRLWiQtssMLgp6EPhWXtaYyStWPSGAb"},"metadata":{"from":"EbP4aYNeTHL6q385GuVpRV"},"type":"0"},"txnMetadata":{"seqNo":2,"txnId":"1ac8aece2a18ced660fef8694b61aac3af08ba875ce3026a160acbc3a3af35fc"},"ver":"1"}\n{"reqSignature":{},"txn":{"data":{"data":{"alias":"Node3","blskey":"3WFpdbg7C5cnLYZwFZevJqhubkFALBfCBBok15GdrKMUhUjGsk3jV6QKj6MZgEubF7oqCafxNdkm7eswgA4sdKTRc82tLGzZBd6vNqU8dupzup6uYUf32KTHTPQbuUM8Yk4QFXjEf2Usu2TJcNkdgpyeUSX42u5LqdDDpNSWUK5deC5","blskey_pop":"QwDeb2CkNSx6r8QC8vGQK3GRv7Yndn84TGNijX8YXHPiagXajyfTjoR87rXUu4G4QLk2cF8NNyqWiYMus1623dELWwx57rLCFqGh7N4ZRbGDRP4fnVcaKg1BcUxQ866Ven4gw8y4N56S5HzxXNBZtLYmhGHvDtk6PFkFwCvxYrNYjh","client_ip":"172.17.0.2","client_port":9706,"node_ip":"172.17.0.2","node_port":9705,"services":["VALIDATOR"]},"dest":"DKVxG2fXXTU8yT5N7hGEbXB3dfdAnYv1JczDUHpmDxya"},"metadata":{"from":"4cU41vWW82ArfxJxHkzXPG"},"type":"0"},"txnMetadata":{"seqNo":3,"txnId":"7e9f355dffa78ed24668f0e0e369fd8c224076571c51e2ea8be5f26479edebe4"},"ver":"1"}\n{"reqSignature":{},"txn":{"data":{"data":{"alias":"Node4","blskey":"2zN3bHM1m4rLz54MJHYSwvqzPchYp8jkHswveCLAEJVcX6Mm1wHQD1SkPYMzUDTZvWvhuE6VNAkK3KxVeEmsanSmvjVkReDeBEMxeDaayjcZjFGPydyey1qxBHmTvAnBKoPydvuTAqx5f7YNNRAdeLmUi99gERUU7TD8KfAa6MpQ9bw","blskey_pop":"RPLagxaR5xdimFzwmzYnz4ZhWtYQEj8iR5ZU53T2gitPCyCHQneUn2Huc4oeLd2B2HzkGnjAff4hWTJT6C7qHYB1Mv2wU5iHHGFWkhnTX9WsEAbunJCV2qcaXScKj4tTfvdDKfLiVuU2av6hbsMztirRze7LvYBkRHV3tGwyCptsrP","client_ip":"172.17.0.2","client_port":9708,"node_ip":"172.17.0.2","node_port":9707,"services":["VALIDATOR"]},"dest":"4PS3EDQ3dW1tci1Bp6543CfuuebjFrg36kLAUcskGfaA"},"metadata":{"from":"TWwCRQRZ2ZHMJFn9TzLp7W"},"type":"0"},"txnMetadata":{"seqNo":4,"txnId":"aa5e817d7cc626170eca175822029339a444eb0ee8f0bd20d3b0b76e566fb008"},"ver":"1"}',
            True,
            "endorser_did",
        )

        assert isinstance(result, tuple)

    @mock.patch.object(CredentialDefinition, "load")
    @mock.patch.object(RevocationRegistryDefinition, "load")
    @mock.patch.object(RevocationRegistryDefinitionPrivate, "load")
    @mock.patch.object(
        IssuerCredRevRecord,
        "query_by_ids",
        return_value=[
            IssuerCredRevRecord(
                state=IssuerCredRevRecord.STATE_REVOKED,
                cred_ex_id="cred_ex_id",
                rev_reg_id="4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
                cred_rev_id="1",
            ),
            IssuerCredRevRecord(
                state=IssuerCredRevRecord.STATE_REVOKED,
                cred_ex_id="cred_ex_id",
                rev_reg_id="4xE68b6S5VRFrKMMG1U95M:5:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
                cred_rev_id="2",
            ),
        ],
    )
    @mock.patch.object(
        RevList,
        "to_native",
        return_value=mock.MagicMock(
            update=mock.MagicMock(return_value=MockRevListEntry())
        ),
    )
    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_sync_wallet_rev_list_with_issuer_cred_rev_records(
        self, mock_handle, *_
    ):
        async with self.profile.session() as session:
            # Matching revocations and rev_list
            mock_handle.fetch = mock.CoroutineMock(
                side_effect=[
                    MockRevRegDefEntry(),
                    MockCredDefEntry(),
                    mock.CoroutineMock(return_value=None),
                ]
            )
            result = await self.registry._sync_wallet_rev_list_with_issuer_cred_rev_records(
                session,
                RevList(
                    issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                    current_accumulator="21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
                    revocation_list=[0, 1, 1, 0],
                    timestamp=1669640864487,
                    rev_reg_def_id="4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
                ),
            )
            assert isinstance(result, RevList)
            # Non-matching revocations and rev_list
            mock_handle.fetch = mock.CoroutineMock(
                side_effect=[
                    MockRevRegDefEntry(),
                    MockCredDefEntry(),
                    mock.CoroutineMock(return_value=None),
                    MockRevListEntry(),
                ]
            )
            mock_handle.replace = mock.CoroutineMock(return_value=None)
            result = await self.registry._sync_wallet_rev_list_with_issuer_cred_rev_records(
                session,
                RevList(
                    issuer_id="CsQY9MGeD3CQP4EyuVFo5m",
                    current_accumulator="21 124C594B6B20E41B681E92B2C43FD165EA9E68BC3C9D63A82C8893124983CAE94 21 124C5341937827427B0A3A32113BD5E64FB7AB39BD3E5ABDD7970874501CA4897 6 5438CB6F442E2F807812FD9DC0C39AFF4A86B1E6766DBB5359E86A4D70401B0F 4 39D1CA5C4716FFC4FE0853C4FF7F081DFD8DF8D2C2CA79705211680AC77BF3A1 6 70504A5493F89C97C225B68310811A41AD9CD889301F238E93C95AD085E84191 4 39582252194D756D5D86D0EED02BF1B95CE12AED2FA5CD3C53260747D891993C",
                    revocation_list=[0, 1, 0, 0],
                    timestamp=1669640864487,
                    rev_reg_def_id="4xE68b6S5VRFrKMMG1U95M:4:4xE68b6S5VRFrKMMG1U95M:3:CL:59232:default:CL_ACCUM:4ae1cc6c-f6bd-486c-8057-88f2ce74e960",
                ),
            )
            assert isinstance(result, RevList)

    async def test_get_schem_info(self):
        result = await self.registry.get_schema_info_by_id(
            self.profile,
            "XduBsoPyEA4szYMy3pZ8De:2:minimal-33279d005748b3cc:1.0",
        )
        assert result.issuer_id == "XduBsoPyEA4szYMy3pZ8De"
        assert result.name == "minimal-33279d005748b3cc"
        assert result.version == "1.0"



"""Custom logging filters for aca-py agent."""

import logging
from contextvars import ContextVar

context_wallet_id: ContextVar[str] = ContextVar("context_wallet_id")


class ContextFilter(logging.Filter):
    """Custom logging Filter to adapt logs with contextual wallet_id."""

    def __init__(self):
        """Initialize an instance of Custom logging.Filter."""
        super().__init__()

    def filter(self, record: logging.LogRecord) -> bool:
        """Filter LogRecords and add wallet_id to them."""
        try:
            wallet_id = context_wallet_id.get()
            record.wallet_id = wallet_id
            return True
        except LookupError:
            record.wallet_id = None
            return True



from unittest import IsolatedAsyncioTestCase, mock

from configargparse import ArgumentTypeError

from .. import argparse
from ..util import BoundedInt, ByteSize


class TestArgParse(IsolatedAsyncioTestCase):
    async def test_groups(self):
        """Test optional argument parsing."""
        parser = argparse.create_argument_parser()

        groups = (
            g for g in argparse.group.get_registered() if g is not argparse.TransportGroup
        )
        argparse.load_argument_groups(parser, *groups)

        parser.parse_args([])

    async def test_transport_settings(self):
        """Test required argument parsing."""

        parser = argparse.create_argument_parser()
        group = argparse.TransportGroup()
        group.add_arguments(parser)

        with mock.patch.object(parser, "exit") as exit_parser:
            parser.parse_args(["-h"])
            exit_parser.assert_called_once()

        result = parser.parse_args(
            [
                "--inbound-transport",
                "http",
                "0.0.0.0",
                "80",
                "--outbound-transport",
                "http",
                "--max-outbound-retry",
                "5",
            ]
        )

        assert result.inbound_transports == [["http", "0.0.0.0", "80"]]
        assert result.outbound_transports == ["http"]

        settings = group.get_settings(result)

        assert settings.get("transport.inbound_configs") == [["http", "0.0.0.0", "80"]]
        assert settings.get("transport.outbound_configs") == ["http"]
        assert result.max_outbound_retry == 5

    async def test_get_genesis_transactions_list_with_ledger_selection(self):
        """Test multiple ledger support related argument parsing."""

        parser = argparse.create_argument_parser()
        group = argparse.LedgerGroup()
        group.add_arguments(parser)

        with mock.patch.object(parser, "exit") as exit_parser:
            parser.parse_args(["-h"])
            exit_parser.assert_called_once()

        result = parser.parse_args(
            [
                "--genesis-transactions-list",
                "./acapy_agent/config/tests/test-ledger-args-no-write.yaml",
            ]
        )
        assert (
            result.genesis_transactions_list
            == "./acapy_agent/config/tests/test-ledger-args-no-write.yaml"
        )
        with self.assertRaises(argparse.ArgsParseError):
            settings = group.get_settings(result)

        result = parser.parse_args(
            [
                "--genesis-transactions-list",
                "./acapy_agent/config/tests/test-ledger-args-no-genesis.yaml",
            ]
        )
        assert (
            result.genesis_transactions_list
            == "./acapy_agent/config/tests/test-ledger-args-no-genesis.yaml"
        )
        with self.assertRaises(argparse.ArgsParseError):
            settings = group.get_settings(result)

        result = parser.parse_args(
            [
                "--genesis-transactions-list",
                "./acapy_agent/config/tests/test-ledger-args.yaml",
            ]
        )
        assert (
            result.genesis_transactions_list
            == "./acapy_agent/config/tests/test-ledger-args.yaml"
        )
        settings = group.get_settings(result)

        assert len(settings.get("ledger.ledger_config_list")) == 3
        assert (
            {
                "id": "sovrinStaging",
                "is_production": True,
                "is_write": True,
                "genesis_file": "/home/indy/ledger/sandbox/pool_transactions_genesis",
                "pool_name": "sovrinStaging",
            }
        ) in settings.get("ledger.ledger_config_list")
        assert (
            {
                "id": "sovrinTest",
                "is_production": False,
                "genesis_url": "http://localhost:9000/genesis",
                "pool_name": "sovrinTest",
            }
        ) in settings.get("ledger.ledger_config_list")

    async def test_upgrade_config(self):
        """Test upgrade command related argument parsing."""

        parser = argparse.create_argument_parser()
        group = argparse.UpgradeGroup()
        group.add_arguments(parser)

        with mock.patch.object(parser, "exit") as exit_parser:
            parser.parse_args(["-h"])
            exit_parser.assert_called_once()

        result = parser.parse_args(
            [
                "--upgrade-config-path",
                "./acapy_agent/config/tests/test-acapy-upgrade-config.yml",
                "--from-version",
                "v0.7.2",
                "--force-upgrade",
            ]
        )

        assert (
            result.upgrade_config_path
            == "./acapy_agent/config/tests/test-acapy-upgrade-config.yml"
        )
        assert result.force_upgrade is True

        settings = group.get_settings(result)

        assert (
            settings.get("upgrade.config_path")
            == "./acapy_agent/config/tests/test-acapy-upgrade-config.yml"
        )

        result = parser.parse_args(
            [
                "--named-tag",
                "test_tag_1",
                "--named-tag",
                "test_tag_2",
                "--force-upgrade",
            ]
        )

        assert result.named_tag == ["test_tag_1", "test_tag_2"]
        assert result.force_upgrade is True

        settings = group.get_settings(result)

        assert settings.get("upgrade.named_tags") == ["test_tag_1", "test_tag_2"]
        assert settings.get("upgrade.force_upgrade") is True

        result = parser.parse_args(
            [
                "--upgrade-config-path",
                "./acapy_agent/config/tests/test-acapy-upgrade-config.yml",
                "--from-version",
                "v0.7.2",
                "--upgrade-all-subwallets",
                "--force-upgrade",
            ]
        )

        assert (
            result.upgrade_config_path
            == "./acapy_agent/config/tests/test-acapy-upgrade-config.yml"
        )
        assert result.force_upgrade is True
        assert result.upgrade_all_subwallets is True

        settings = group.get_settings(result)

        assert (
            settings.get("upgrade.config_path")
            == "./acapy_agent/config/tests/test-acapy-upgrade-config.yml"
        )
        assert settings.get("upgrade.force_upgrade") is True
        assert settings.get("upgrade.upgrade_all_subwallets") is True

        result = parser.parse_args(
            [
                "--named-tag",
                "fix_issue_rev_reg",
                "--upgrade-subwallet",
                "test_wallet_id_1",
                "--upgrade-subwallet",
                "test_wallet_id_2",
                "--force-upgrade",
            ]
        )

        assert result.named_tag == ["fix_issue_rev_reg"]
        assert result.force_upgrade is True
        assert result.upgrade_subwallet == ["test_wallet_id_1", "test_wallet_id_2"]

        settings = group.get_settings(result)
        assert settings.get("upgrade.named_tags") == ["fix_issue_rev_reg"]
        assert settings.get("upgrade.force_upgrade") is True
        assert settings.get("upgrade.upgrade_subwallets") == [
            "test_wallet_id_1",
            "test_wallet_id_2",
        ]

    async def test_outbound_is_required(self):
        """Test that either -ot or -oq are required"""
        parser = argparse.create_argument_parser()
        group = argparse.TransportGroup()
        group.add_arguments(parser)

        result = parser.parse_args(
            [
                "--inbound-transport",
                "http",
                "0.0.0.0",
                "80",
            ]
        )

        with self.assertRaises(argparse.ArgsParseError):
            settings = group.get_settings(result)

    async def test_general_settings_file(self):
        """Test file argument parsing."""

        parser = argparse.create_argument_parser()
        group = argparse.GeneralGroup()
        group.add_arguments(parser)

        with mock.patch.object(parser, "exit") as exit_parser:
            parser.parse_args(["-h"])
            exit_parser.assert_called_once()

        result = parser.parse_args(
            [
                "--arg-file",
                "./acapy_agent/config/tests/test-general-args.yaml",
            ]
        )

        assert result.external_plugins == ["foo"]
        assert result.storage_type == "bar"

        settings = group.get_settings(result)

        assert settings.get("external_plugins") == ["foo"]
        assert settings.get("storage_type") == "bar"

    async def test_plugin_config_file(self):
        """Test file argument parsing."""

        parser = argparse.create_argument_parser()
        group = argparse.GeneralGroup()
        group.add_arguments(parser)

        result = parser.parse_args(
            [
                "--endpoint",
                "localhost",
                "--plugin-config",
                "./acapy_agent/config/tests/test_plugins_config.yaml",
            ]
        )

        assert (
            result.plugin_config == "./acapy_agent/config/tests/test_plugins_config.yaml"
        )

        settings = group.get_settings(result)

        assert settings.get("plugin_config").get("mock_resolver") == {
            "methods": ["sov", "btcr"]
        }

    async def test_transport_settings_file(self):
        """Test file argument parsing."""

        parser = argparse.create_argument_parser()
        group = argparse.GeneralGroup()
        group.add_arguments(parser)
        group = argparse.TransportGroup()
        group.add_arguments(parser)

        with mock.patch.object(parser, "exit") as exit_parser:
            parser.parse_args(["-h"])
            exit_parser.assert_called_once()

        result = parser.parse_args(
            [
                "--arg-file",
                "./acapy_agent/config/tests/test-transport-args.yaml",
            ]
        )
        # no asserts, just testing that the parser doesn't fail

    async def test_multitenancy_settings(self):
        """Test required argument parsing."""

        parser = argparse.create_argument_parser()
        group = argparse.MultitenantGroup()
        group.add_arguments(parser)

        result = parser.parse_args(
            [
                "--multitenant",
                "--jwt-secret",
                "secret",
                "--multitenancy-config",
                '{"wallet_type":"askar","wallet_name":"test", "cache_size": 10}',
                "--base-wallet-routes",
                "/my_route",
            ]
        )

        settings = group.get_settings(result)

        assert settings.get("multitenant.enabled") is True
        assert settings.get("multitenant.jwt_secret") == "secret"
        assert settings.get("multitenant.wallet_type") == "askar"
        assert settings.get("multitenant.wallet_name") == "test"
        assert settings.get("multitenant.base_wallet_routes") == ["/my_route"]

        result = parser.parse_args(
            [
                "--multitenant",
                "--jwt-secret",
                "secret",
                "--multitenancy-config",
                "wallet_type=askar",
                "wallet_name=test",
                "cache_size=10",
                "--base-wallet-routes",
                "/my_route",
            ]
        )

        settings = group.get_settings(result)

        assert settings.get("multitenant.enabled") is True
        assert settings.get("multitenant.jwt_secret") == "secret"
        assert settings.get("multitenant.wallet_type") == "askar"
        assert settings.get("multitenant.wallet_name") == "test"
        assert settings.get("multitenant.base_wallet_routes") == ["/my_route"]

    async def test_endorser_settings(self):
        """Test required argument parsing."""

        parser = argparse.create_argument_parser()
        group = argparse.EndorsementGroup()
        group.add_arguments(parser)

        result = parser.parse_args(
            [
                "--endorser-protocol-role",
                argparse.ENDORSER_AUTHOR,
                "--endorser-public-did",
                "did:sov:12345",
            ]
        )

        settings = group.get_settings(result)

        assert settings.get("endorser.author") is True
        assert settings.get("endorser.endorser") is False
        assert settings.get("endorser.endorser_public_did") == "did:sov:12345"
        assert settings.get("endorser.auto_endorse") is False

    async def test_logging(self):
        """Test logging."""

        parser = argparse.create_argument_parser()
        group = argparse.LoggingGroup()
        group.add_arguments(parser)

        result = parser.parse_args(
            [
                "--log-file",
                "test_file.log",
                "--log-level",
                "INFO",
            ]
        )

        settings = group.get_settings(result)

        assert settings.get("log.file") == "test_file.log"
        assert settings.get("log.level") == "INFO"

    async def test_error_raised_when_multitenancy_used_and_no_jwt_provided(self):
        """Test that error is raised if no jwt_secret is provided with multitenancy."""

        parser = argparse.create_argument_parser()
        group = argparse.MultitenantGroup()
        group.add_arguments(parser)

        result = parser.parse_args(
            [
                "--multitenant",
                "--multitenancy-config",
                '{"wallet_type":"askar","wallet_name":"test"}',
            ]
        )

        with self.assertRaises(argparse.ArgsParseError):
            settings = group.get_settings(result)

    def test_bytesize(self):
        bs = ByteSize()
        with self.assertRaises(ArgumentTypeError):
            bs(None)
        with self.assertRaises(ArgumentTypeError):
            bs("")
        with self.assertRaises(ArgumentTypeError):
            bs("a")
        with self.assertRaises(ArgumentTypeError):
            bs("1.5")
        with self.assertRaises(ArgumentTypeError):
            bs("-1")
        assert bs("101") == 101
        assert bs("101b") == 101
        assert bs("101KB") == 103424
        assert bs("2M") == 2097152
        assert bs("1G") == 1073741824
        assert bs("1t") == 1099511627776

        bs = ByteSize(min=10)
        with self.assertRaises(ArgumentTypeError):
            bs("5")
        assert bs("12") == 12

        bs = ByteSize(max=10)
        with self.assertRaises(ArgumentTypeError):
            bs("15")
        assert bs("10") == 10

        assert repr(bs) == "bytes"

    def test_bounded_int(self):
        bounded = BoundedInt()
        with self.assertRaises(ArgumentTypeError):
            bounded(None)
        with self.assertRaises(ArgumentTypeError):
            bounded("")
        with self.assertRaises(ArgumentTypeError):
            bounded("a")
        with self.assertRaises(ArgumentTypeError):
            bounded("1.5")
        assert bounded("101") == 101
        assert bounded("-99") == -99

        bounded = BoundedInt(min=10)
        with self.assertRaises(ArgumentTypeError):
            bounded("5")
        assert bounded("12") == 12

        bounded = BoundedInt(max=10)
        with self.assertRaises(ArgumentTypeError):
            bounded("15")
        assert bounded("10") == 10

        assert repr(bounded) == "integer"

    async def test_mediation_x_clear_and_default(self):
        parser = argparse.create_argument_parser()
        group = argparse.MediationGroup()
        group.add_arguments(parser)

        with self.assertRaises(argparse.ArgsParseError):
            args = parser.parse_args(
                ["--clear-default-mediator", "--default-mediator-id", "asdf"]
            )
            group.get_settings(args)

    def test_plugin_config_value_parsing(self):
        required_args = ["-e", "http://localhost:3000"]
        parser = argparse.create_argument_parser()
        group = argparse.GeneralGroup()
        group.add_arguments(parser)
        args = parser.parse_args(
            [
                *required_args,
                "--plugin-config-value",
                "a.b.c=test",
                "a.b.d=one",
                "--plugin-config-value",
                "x.y.z=value",
                "--plugin-config-value",
                "a_dict={key: value}",
                "--plugin-config-value",
                "a_list=[one, two]",
            ]
        )
        settings = group.get_settings(args)

        assert settings["plugin_config"]["a"]["b"]["c"] == "test"
        assert settings["plugin_config"]["a"]["b"]["d"] == "one"
        assert settings["plugin_config"]["x"]["y"]["z"] == "value"
        assert settings["plugin_config"]["a_dict"] == {"key": "value"}
        assert settings["plugin_config"]["a_list"] == ["one", "two"]

    async def test_wallet_key_derivation_method_value_parsing(self):
        key_derivation_method = "key_derivation_method"
        parser = argparse.create_argument_parser()
        group = argparse.WalletGroup()
        group.add_arguments(parser)

        result = parser.parse_args(
            ["--wallet-key-derivation-method", key_derivation_method, "--wallet-test"]
        )

        settings = group.get_settings(result)

        assert settings.get("wallet.key_derivation_method") == key_derivation_method

    async def test_wallet_key_value_parsing(self):
        key_value = "some_key_value"
        parser = argparse.create_argument_parser()
        group = argparse.WalletGroup()
        group.add_arguments(parser)

        result = parser.parse_args(
            [
                "--wallet-key",
                key_value,
                "--wallet-test",
            ]
        )

        settings = group.get_settings(result)

        assert settings.get("wallet.key") == key_value

    async def test_discover_features_args(self):
        """Test discover features support related argument parsing."""

        parser = argparse.create_argument_parser()
        group = argparse.DiscoverFeaturesGroup()
        group.add_arguments(parser)

        with mock.patch.object(parser, "exit") as exit_parser:
            parser.parse_args(["-h"])
            exit_parser.assert_called_once()

        result = parser.parse_args(
            args=(
                "--auto-disclose-features --disclose-features-list"
                " ./acapy_agent/config/tests/test_disclose_features_list.yaml"
            )
        )

        assert result.auto_disclose_features
        assert (
            result.disclose_features_list
            == "./acapy_agent/config/tests/test_disclose_features_list.yaml"
        )

        settings = group.get_settings(result)

        assert settings.get("auto_disclose_features")
        assert (["test_protocol_1", "test_protocol_2"]) == settings.get(
            "disclose_protocol_list"
        )
        assert (["test_goal_code_1", "test_goal_code_2"]) == settings.get(
            "disclose_goal_code_list"
        )

    def test_universal_resolver(self):
        """Test universal resolver flags."""
        parser = argparse.create_argument_parser()
        group = argparse.GeneralGroup()
        group.add_arguments(parser)

        result = parser.parse_args(["-e", "test", "--universal-resolver"])
        settings = group.get_settings(result)
        endpoint = settings.get("resolver.universal")
        assert endpoint
        assert endpoint == "DEFAULT"

        result = parser.parse_args(
            ["-e", "test", "--universal-resolver", "https://example.com"]
        )
        settings = group.get_settings(result)
        endpoint = settings.get("resolver.universal")
        assert endpoint
        assert endpoint == "https://example.com"

        result = parser.parse_args(
            [
                "-e",
                "test",
                "--universal-resolver",
                "https://example.com",
                "--universal-resolver-regex",
                "regex",
            ]
        )
        settings = group.get_settings(result)
        endpoint = settings.get("resolver.universal")
        assert endpoint
        assert endpoint == "https://example.com"
        supported_regex = settings.get("resolver.universal.supported")
        assert supported_regex
        assert supported_regex == ["regex"]

        result = parser.parse_args(["-e", "test", "--universal-resolver-regex", "regex"])
        with self.assertRaises(argparse.ArgsParseError):
            group.get_settings(result)



import pytest

from ...storage.record import StorageRecord


@pytest.fixture
def record_factory():
    def _test_record(tags={}):
        return StorageRecord(type="TYPE", value="TEST", tags=tags)

    yield _test_record


@pytest.fixture
def missing():
    yield StorageRecord(type="__MISSING__", value="000000000")



from unittest import IsolatedAsyncioTestCase, TestCase, mock

from .....didcomm_prefix import DIDCommPrefix
from ...message_types import TRANSACTION_JOB_TO_SEND
from ..transaction_job_to_send import TransactionJobToSend


class TestConfig:
    test_job = "TRANSACTION_AUTHOR"


class TestTransactionJobToSend(TestCase, TestConfig):
    def setUp(self):
        self.transaction_job_to_send = TransactionJobToSend(job=self.test_job)

    def test_init(self):
        """Test initialization."""
        assert self.transaction_job_to_send.job == self.test_job

    def test_type(self):
        """Test type."""
        assert self.transaction_job_to_send._type == DIDCommPrefix.qualify_current(
            TRANSACTION_JOB_TO_SEND
        )

    @mock.patch(
        "acapy_agent.protocols.endorse_transaction.v1_0.messages."
        "transaction_job_to_send.TransactionJobToSendSchema.load"
    )
    def test_deserialize(self, mock_transaction_job_to_send_schema_load):
        """
        Test deserialization.
        """
        obj = self.transaction_job_to_send

        transaction_job_to_send = TransactionJobToSend.deserialize(obj)
        mock_transaction_job_to_send_schema_load.assert_called_once_with(obj)

        assert (
            transaction_job_to_send
            is mock_transaction_job_to_send_schema_load.return_value
        )

    @mock.patch(
        "acapy_agent.protocols.endorse_transaction.v1_0.messages."
        "transaction_job_to_send.TransactionJobToSendSchema.dump"
    )
    def test_serialize(self, mock_transaction_job_to_send_schema_dump):
        """
        Test serialization.
        """
        transaction_job_to_send_dict = self.transaction_job_to_send.serialize()
        mock_transaction_job_to_send_schema_dump.assert_called_once_with(
            self.transaction_job_to_send
        )

        assert (
            transaction_job_to_send_dict
            is mock_transaction_job_to_send_schema_dump.return_value
        )


class TestTransactionJobToSendSchema(IsolatedAsyncioTestCase, TestConfig):
    """Test transaction job to send schema."""

    async def test_make_model(self):
        transaction_job_to_send = TransactionJobToSend(job=self.test_job)
        data = transaction_job_to_send.serialize()
        model_instance = TransactionJobToSend.deserialize(data)
        assert type(model_instance) is type(transaction_job_to_send)



"""DataIntegrityProof."""

from typing import List, Optional

from marshmallow import INCLUDE, fields

from ....messaging.models.base import BaseModel, BaseModelSchema
from .proof import DataIntegrityProof, DataIntegrityProofSchema


class ProblemDetails(BaseModel):
    """ProblemDetails model."""

    class Meta:
        """ProblemDetails metadata."""

        schema_class = "ProblemDetailsSchema"

    def __init__(
        self,
        type: Optional[str] = None,
        title: Optional[str] = None,
        detail: Optional[str] = None,
    ) -> None:
        """Initialize the ProblemDetails instance."""

        self.type = type
        self.title = title
        self.detail = detail


class ProblemDetailsSchema(BaseModelSchema):
    """ProblemDetails schema.

    Based on https://www.w3.org/TR/vc-data-model-2.0/#problem-details.

    """

    class Meta:
        """Accept parameter overload."""

        unknown = INCLUDE
        model_class = ProblemDetails

    type = fields.Str(
        required=True,
        metadata={
            "example": "https://w3id.org/security#PROOF_VERIFICATION_ERROR",
        },
    )

    title = fields.Str(
        required=False,
        metadata={},
    )

    detail = fields.Str(
        required=False,
        metadata={},
    )


class DataIntegrityVerificationResult(BaseModel):
    """Data Integrity Verification Result model."""

    class Meta:
        """DataIntegrityVerificationResult metadata."""

        schema_class = "DataIntegrityVerificationResultSchema"

    def __init__(
        self,
        verified: Optional[bool] = None,
        proof: Optional[DataIntegrityProof] = None,
        problem_details: Optional[List[ProblemDetails]] = None,
    ) -> None:
        """Initialize the DataIntegrityVerificationResult instance."""

        self.verified = verified
        self.proof = proof
        self.problem_details = problem_details


class DataIntegrityVerificationResultSchema(BaseModelSchema):
    """DataIntegrityVerificationResult schema."""

    class Meta:
        """Accept parameter overload."""

        unknown = INCLUDE
        model_class = DataIntegrityVerificationResult

    verified = fields.Bool(
        required=True,
        metadata={
            "example": False,
        },
    )

    proof = fields.Nested(
        DataIntegrityProofSchema(),
        required=True,
        metadata={},
    )

    problem_details = fields.List(
        fields.Nested(ProblemDetailsSchema()),
        data_key="problemDetails",
        required=True,
        metadata={},
    )


class DataIntegrityVerificationResponse(BaseModel):
    """Data Integrity Verification Response model."""

    class Meta:
        """DataIntegrityVerificationResponse metadata."""

        schema_class = "DataIntegrityVerificationResponseSchema"

    def __init__(
        self,
        verified: Optional[bool] = None,
        verified_document: Optional[dict] = None,
        results: Optional[List[DataIntegrityVerificationResult]] = None,
    ) -> None:
        """Initialize the DataIntegrityVerificationResponse instance."""

        self.verified = verified
        self.verified_document = verified_document
        self.results = results


class DataIntegrityVerificationResponseSchema(BaseModelSchema):
    """DataIntegrityVerificationResponse schema."""

    class Meta:
        """Accept parameter overload."""

        unknown = INCLUDE
        model_class = DataIntegrityVerificationResponse

    verified = fields.Bool(
        required=True,
        metadata={
            "example": False,
        },
    )

    verified_document = fields.Dict(
        data_key="verifiedDocument",
        required=False,
        metadata={},
    )

    results = fields.List(
        fields.Nested(DataIntegrityVerificationResultSchema()),
        required=False,
        metadata={},
    )



"""DID exchange response handler under RFC 23."""

from .....messaging.base_handler import BaseHandler, BaseResponder, RequestContext
from ....trustping.v1_0.messages.ping import Ping
from ..manager import DIDXManager, DIDXManagerError
from ..messages.response import DIDXResponse


class DIDXResponseHandler(BaseHandler):
    """Handler class for DID exchange response message under RFC 23."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle DID exchange response under RFC 23.

        Args:
            context: Request context
            responder: Responder callback
        """
        self._logger.debug("DIDXResponseHandler called with context %s", context)
        assert isinstance(context.message, DIDXResponse)

        profile = context.profile
        mgr = DIDXManager(profile)
        try:
            conn_rec = await mgr.accept_response(context.message, context.message_receipt)
        except DIDXManagerError as e:
            report, targets = await mgr.manager_error_to_problem_report(
                e, context.message, context.message_receipt
            )
            if report and targets:
                await responder.send_reply(
                    message=report,
                    target_list=targets,
                )
            return

        # send trust ping in response
        if context.settings.get("auto_ping_connection"):
            await responder.send(Ping(), connection_id=conn_rec.connection_id)



"""Handler for incoming invitation messages."""

from typing import Optional

from .....messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    HandlerException,
    RequestContext,
)
from ..base_service import BaseIntroductionService
from ..messages.invitation import Invitation as IntroInvitation


class InvitationHandler(BaseHandler):
    """Handler for incoming invitation messages."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler implementation."""
        self._logger.debug("InvitationHandler called with context %s", context)
        assert isinstance(context.message, IntroInvitation)

        if not context.connection_ready:
            raise HandlerException("No connection established for invitation message")

        service: Optional[BaseIntroductionService] = context.inject_or(
            BaseIntroductionService
        )
        if service:
            async with context.profile.session() as session:
                await service.return_invitation(
                    context.connection_record.connection_id,
                    context.message,
                    session,
                    responder.send,
                )
        else:
            raise HandlerException(
                "Cannot handle Invitation message with no introduction service"
            )



"""Admin routes for presentations."""

from marshmallow import fields

from ...messaging.models.openapi import OpenAPISchema
from ...messaging.valid import INT_EPOCH_EXAMPLE, INT_EPOCH_VALIDATE


class IndyRequestedCredsRequestedAttrSchema(OpenAPISchema):
    """Schema for requested attributes within indy requested credentials structure."""

    cred_id = fields.Str(
        required=True,
        metadata={
            "example": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
            "description": (
                "Wallet credential identifier (typically but not necessarily a UUID)"
            ),
        },
    )
    revealed = fields.Bool(
        dump_default=True,
        metadata={"description": "Whether to reveal attribute in proof (default true)"},
    )


class IndyRequestedCredsRequestedPredSchema(OpenAPISchema):
    """Schema for requested predicates within indy requested credentials structure."""

    cred_id = fields.Str(
        required=True,
        metadata={
            "description": (
                "Wallet credential identifier (typically but not necessarily a UUID)"
            ),
            "example": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
        },
    )
    timestamp = fields.Int(
        required=False,
        validate=INT_EPOCH_VALIDATE,
        metadata={
            "description": "Epoch timestamp of interest for non-revocation proof",
            "strict": True,
            "example": INT_EPOCH_EXAMPLE,
        },
    )



from unittest import TestCase

from ..indy import NonRevocationInterval

FROM = 1000000000
TO = 1234567890

INTERVAL_FROM = NonRevocationInterval(fro=FROM)
INTERVAL_TO = NonRevocationInterval(to=TO)
INTERVAL = NonRevocationInterval(fro=FROM, to=TO)


class TestInterval(TestCase):
    """Non-revocation interval tests"""

    def test_serde(self):
        """Test serialization and deserialization."""
        for interval in (INTERVAL_FROM, INTERVAL_TO, INTERVAL):
            non_revo_dict = interval.serialize()
            assert non_revo_dict.get("from") == interval.fro
            assert non_revo_dict.get("to") == interval.to

            model = NonRevocationInterval.deserialize(non_revo_dict)
            assert model.fro == interval.fro and model.to == interval.to
            assert model.timestamp()

    def test_covers(self):
        """Test spanning check."""
        assert INTERVAL_FROM.covers(FROM)
        assert INTERVAL_FROM.covers(TO)
        assert INTERVAL_FROM.covers()

        assert INTERVAL_TO.covers(FROM)
        assert INTERVAL_TO.covers(TO)
        assert not INTERVAL_TO.covers()

        assert INTERVAL.covers(FROM)
        assert INTERVAL.covers(TO)
        assert not INTERVAL.covers()



from unittest import IsolatedAsyncioTestCase, TestCase, mock

from .....didcomm_prefix import DIDCommPrefix
from ...message_types import BASIC_MESSAGE, PROTOCOL_PACKAGE
from ..basicmessage import BasicMessage


class TestBasicMessage(TestCase):
    def setUp(self):
        self.test_content = "hello"
        self.test_message = BasicMessage(content=self.test_content)

    def test_init(self):
        """Test initialization."""
        assert self.test_message.content == self.test_content
        assert self.test_message.sent_time

    def test_type(self):
        """Test type."""
        assert self.test_message._type == DIDCommPrefix.qualify_current(BASIC_MESSAGE)

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.basicmessage.BasicMessageSchema.load")
    def test_deserialize(self, mock_basic_message_schema_load):
        """
        Test deserialization.
        """
        obj = {"obj": "obj"}

        msg = BasicMessage.deserialize(obj)
        mock_basic_message_schema_load.assert_called_once_with(obj)

        assert msg is mock_basic_message_schema_load.return_value

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.basicmessage.BasicMessageSchema.dump")
    def test_serialize(self, mock_basic_message_schema_load):
        """
        Test serialization.
        """
        msg_dict = self.test_message.serialize()
        mock_basic_message_schema_load.assert_called_once_with(self.test_message)

        assert msg_dict is mock_basic_message_schema_load.return_value


class TestBasicMessageSchema(IsolatedAsyncioTestCase):
    """Test basic message schema."""

    async def test_make_model(self):
        basic_message = BasicMessage(content="hello")
        data = basic_message.serialize()
        model_instance = BasicMessage.deserialize(data)
        assert type(model_instance) is type(basic_message)



from unittest import IsolatedAsyncioTestCase

from ..did_posture import DIDPosture


class TestDIDPosture(IsolatedAsyncioTestCase):
    async def test_did_posture(self):
        assert DIDPosture.PUBLIC is DIDPosture.get("PUBLIC")
        assert DIDPosture.POSTED is DIDPosture.get("POSTED")
        assert DIDPosture.WALLET_ONLY is DIDPosture.get("WALLET_ONLY")
        assert DIDPosture.get("no-such-type") is None
        assert DIDPosture.get(None) is None

        assert DIDPosture.PUBLIC is DIDPosture.get({"public": True})
        assert DIDPosture.PUBLIC is DIDPosture.get({"public": True, "posted": True})
        assert DIDPosture.POSTED is DIDPosture.get({"public": False, "posted": True})
        assert DIDPosture.WALLET_ONLY is DIDPosture.get(
            {"public": False, "posted": False}
        )
        assert DIDPosture.WALLET_ONLY is DIDPosture.get({})

        postures = list(DIDPosture)
        postures.sort(key=lambda p: p.ordinal)
        assert postures == [
            DIDPosture.PUBLIC,
            DIDPosture.POSTED,
            DIDPosture.WALLET_ONLY,
        ]



"""Enum representing captured send status of outbound messages."""

from enum import Enum

OUTBOUND_STATUS_PREFIX = "acapy::outbound-message::"


class OutboundSendStatus(Enum):
    """Send status of outbound messages."""

    # Could directly send the message to the connection over active session
    SENT_TO_SESSION = "sent_to_session"

    # Message is sent to external queue. We don't know how it will process the queue
    SENT_TO_EXTERNAL_QUEUE = "sent_to_external_queue"

    # Message is queued for delivery using outbound transport (recipient has endpoint)
    QUEUED_FOR_DELIVERY = "queued_for_delivery"

    # No endpoint available.
    # Need to wait for the recipient to connect with return routing for delivery
    WAITING_FOR_PICKUP = "waiting_for_pickup"

    # No endpoint available, and no internal queue for messages.
    UNDELIVERABLE = "undeliverable"

    @property
    def topic(self):
        """Return an event topic associated with a given status."""
        return f"{OUTBOUND_STATUS_PREFIX}{self.value}"



"""Endorser utilities."""

from ....connections.models.conn_record import ConnRecord
from ....core.profile import Profile


def is_author_role(profile: Profile):
    """Check if agent is running in author mode."""
    return profile.settings.get_value("endorser.author")


async def get_endorser_connection_id(profile: Profile):
    """Determine default endorser connection for author."""
    if not is_author_role(profile):
        return None

    endorser_alias = profile.settings.get_value("endorser.endorser_alias")
    if not endorser_alias:
        return None
    try:
        async with profile.session() as session:
            connection_records = await ConnRecord.retrieve_by_alias(
                session, endorser_alias
            )
            connection_id = connection_records[0].connection_id
            return connection_id
    except Exception:
        return None



"""Key pair based on base wallet interface."""

from typing import List, Optional, Union

from base58 import b58encode

from ....core.profile import Profile
from ....utils.multiformats import multibase, multicodec
from ....wallet.base import BaseWallet
from ....wallet.key_type import KeyType
from ....wallet.util import b58_to_bytes
from ..error import LinkedDataProofException
from .key_pair import KeyPair


class WalletKeyPair(KeyPair):
    """Base wallet key pair."""

    def __init__(
        self,
        *,
        profile: Profile,
        key_type: KeyType,
        public_key_base58: Optional[str] = None,
    ) -> None:
        """Initialize new WalletKeyPair instance."""
        super().__init__()
        self.profile = profile
        self.key_type = key_type
        self.public_key_base58 = public_key_base58

    async def sign(self, message: Union[List[bytes], bytes]) -> bytes:
        """Sign message using wallet."""
        if not self.public_key_base58:
            raise LinkedDataProofException(
                "Unable to sign message with WalletKey: No key to sign with"
            )
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            return await wallet.sign_message(
                message=message,
                from_verkey=self.public_key_base58,
            )

    async def verify(self, message: Union[List[bytes], bytes], signature: bytes) -> bool:
        """Verify message against signature using wallet."""
        if not self.public_key_base58:
            raise LinkedDataProofException(
                "Unable to verify message with key pair: No key to verify with"
            )

        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            return await wallet.verify_message(
                message=message,
                signature=signature,
                from_verkey=self.public_key_base58,
                key_type=self.key_type,
            )

    def from_verification_method(self, verification_method: dict) -> "WalletKeyPair":
        """Create new WalletKeyPair from public key in verification method."""
        if "publicKeyBase58" in verification_method:
            key_material = verification_method["publicKeyBase58"]
        elif "sec:publicKeyMultibase" in verification_method:
            # verification_method is framed
            _, raw_key = multicodec.unwrap(
                multibase.decode(verification_method["sec:publicKeyMultibase"]["@value"])
            )
            key_material = b58encode(raw_key).decode()
        else:
            raise LinkedDataProofException(
                f"Unrecognized verification method type: {verification_method}"
            )

        return WalletKeyPair(
            profile=self.profile,
            key_type=self.key_type,
            public_key_base58=key_material,
        )

    @property
    def public_key(self) -> Optional[bytes]:
        """Getter for public key."""
        return b58_to_bytes(self.public_key_base58)

    @property
    def has_public_key(self) -> bool:
        """Whether key pair has public key."""
        return self.public_key_base58 is not None



from unittest import TestCase

from ..validation_result import PresentationVerificationResult


class TestValidationResult(TestCase):
    def test_properties(self):
        result = PresentationVerificationResult(verified=True)
        result2 = PresentationVerificationResult(verified=True)

        assert result.__repr__()
        assert result == result2
        assert result != 10



"""Credential request message handler."""

from .....core.oob_processor import OobMessageProcessor
from .....indy.issuer import IndyIssuerError
from .....ledger.error import LedgerError
from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.models.base import BaseModelError
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageError, StorageNotFoundError
from .....utils.tracing import get_timer, trace_event
from .. import problem_report_for_record
from ..manager import CredentialManager, CredentialManagerError
from ..messages.credential_problem_report import ProblemReportReason
from ..messages.credential_request import CredentialRequest


class CredentialRequestHandler(BaseHandler):
    """Message handler class for credential requests."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for credential requests.

        Args:
            context: request context
            responder: responder callback

        """
        r_time = get_timer()
        profile = context.profile
        self._logger.debug("CredentialRequestHandler called with context %s", context)
        assert isinstance(context.message, CredentialRequest)
        self._logger.debug(
            "Received credential request message: %s",
            context.message.serialize(as_string=True),
        )

        # If connection is present it must be ready for use
        if context.connection_record and not context.connection_ready:
            raise HandlerException("Connection used for credential request not ready")

        # Find associated oob record. If the credential offer was created as an oob
        # attachment the presentation exchange record won't have a connection id (yet)
        oob_processor = context.inject(OobMessageProcessor)
        oob_record = await oob_processor.find_oob_record_for_inbound_message(context)

        # Either connection or oob context must be present
        if not context.connection_record and not oob_record:
            raise HandlerException(
                "No connection or associated connectionless exchange found for credential"
                " request"
            )

        credential_manager = CredentialManager(profile)
        try:
            cred_ex_record = await credential_manager.receive_request(
                context.message, context.connection_record, oob_record
            )  # mgr only finds, saves record: on exception, saving state null is hopeless
        except StorageNotFoundError:
            # issue a problem report...
            cred_ex_record = None
            thread_id = context.message._thread_id
            await responder.send_reply(
                problem_report_for_record(
                    None,
                    ProblemReportReason.RECORD_NOT_FOUND.value,
                    thread_id=thread_id,
                )
            )
        r_time = trace_event(
            context.settings,
            context.message,
            outcome="CredentialRequestHandler.handle.END",
            perf_counter=r_time,
        )

        # If auto_issue is enabled, respond immediately
        if cred_ex_record and cred_ex_record.auto_issue:
            if (
                cred_ex_record.credential_proposal_dict
                and cred_ex_record.credential_proposal_dict.credential_proposal
            ):
                credential_issue_message = None
                try:
                    (
                        cred_ex_record,
                        credential_issue_message,
                    ) = await credential_manager.issue_credential(
                        cred_ex_record=cred_ex_record,
                        comment=context.message.comment,
                    )
                    await responder.send_reply(credential_issue_message)
                except (
                    BaseModelError,
                    CredentialManagerError,
                    IndyIssuerError,
                    LedgerError,
                    StorageError,
                ) as err:
                    self._logger.exception("Error responding to credential request")
                    if cred_ex_record:
                        async with profile.session() as session:
                            await cred_ex_record.save_error_state(
                                session,
                                reason=err.roll_up,  # us: be specific
                            )
                        await responder.send_reply(  # them: be vague
                            problem_report_for_record(
                                cred_ex_record,
                                ProblemReportReason.ISSUANCE_ABANDONED.value,
                            )
                        )

                trace_event(
                    context.settings,
                    credential_issue_message,
                    outcome="CredentialRequestHandler.issue.END",
                    perf_counter=r_time,
                )
            else:
                self._logger.warning(
                    "Operation set for auto-issue but credential exchange record "
                    f"{cred_ex_record.credential_exchange_id} "
                    "has no attribute values"
                )



"""Action menu utility methods."""

from ....admin.request_context import AdminRequestContext
from ....storage.base import BaseStorage, StorageNotFoundError, StorageRecord
from .messages.menu import Menu

MENU_RECORD_TYPE = "connection_action_menu"


async def retrieve_connection_menu(
    connection_id: str, context: AdminRequestContext
) -> Menu:
    """Retrieve the previously-received action menu."""
    async with context.profile.session() as session:
        storage = session.inject(BaseStorage)
        try:
            record = await storage.find_record(
                MENU_RECORD_TYPE, {"connection_id": connection_id}
            )
        except StorageNotFoundError:
            record = None
    return Menu.from_json(record.value) if record else None


async def save_connection_menu(
    menu: Menu, connection_id: str, context: AdminRequestContext
):
    """Save a received action menu."""
    async with context.profile.session() as session:
        storage = session.inject(BaseStorage)
        try:
            record = await storage.find_record(
                MENU_RECORD_TYPE, {"connection_id": connection_id}
            )
        except StorageNotFoundError:
            if menu:
                record = StorageRecord(
                    type=MENU_RECORD_TYPE,
                    value=menu.to_json(),
                    tags={"connection_id": connection_id},
                )
                await storage.add_record(record)
        else:
            if menu:
                await storage.update_record(
                    record, menu.to_json(), {"connection_id": connection_id}
                )
            else:
                await storage.delete_record(record)

    await context.profile.notify(
        "acapy::actionmenu::received",
        {
            "connection_id": connection_id,
            "menu": menu.serialize() if menu else None,
        },
    )



"""Represents a Handshake Reuse message under RFC 0434."""

from typing import Optional, Text

from marshmallow import EXCLUDE, ValidationError, fields, pre_dump

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import DEFAULT_VERSION, MESSAGE_REUSE, PROTOCOL_PACKAGE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.reuse_handler.HandshakeReuseMessageHandler"


class HandshakeReuse(AgentMessage):
    """Class representing a Handshake Reuse message."""

    class Meta:
        """Metadata for Handshake Reuse message."""

        handler_class = HANDLER_CLASS
        message_type = MESSAGE_REUSE
        schema_class = "HandshakeReuseSchema"

    def __init__(
        self,
        version: str = DEFAULT_VERSION,
        msg_type: Optional[Text] = None,
        **kwargs,
    ):
        """Initialize Handshake Reuse message object."""
        super().__init__(_type=msg_type, _version=version, **kwargs)


class HandshakeReuseSchema(AgentMessageSchema):
    """Handshake Reuse schema class."""

    class Meta:
        """Handshake Reuse schema metadata."""

        model_class = HandshakeReuse
        unknown = EXCLUDE

    _type = fields.Str(
        data_key="@type",
        required=False,
        metadata={
            "description": "Message type",
            "example": "https://didcomm.org/my-family/1.0/my-message-type",
        },
    )

    @pre_dump
    def check_thread_deco(self, obj, **kwargs):
        """Thread decorator, and its thid and pthid, are mandatory."""
        if not obj._decorators.to_dict().get("~thread", {}).keys() >= {"thid", "pthid"}:
            raise ValidationError("Missing required field(s) in thread decorator")
        return obj



from unittest import IsolatedAsyncioTestCase

from ......connections.models import connection_target
from ......connections.models.diddoc import DIDDoc, PublicKey, PublicKeyType, Service
from ......messaging.decorators.attach_decorator import AttachDecorator
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......tests import mock
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ......wallet.base import BaseWallet
from ......wallet.did_method import SOV, DIDMethods
from ......wallet.key_type import ED25519
from .....trustping.v1_0.messages.ping import Ping
from ...handlers import response_handler as test_module
from ...manager import DIDXManagerError
from ...messages.problem_report import DIDXProblemReport, ProblemReportReason
from ...messages.response import DIDXResponse

TEST_DID = "55GkHamhTU1ZbTbV2ab9DE"
TEST_VERKEY = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
TEST_LABEL = "Label"
TEST_ENDPOINT = "http://localhost"
TEST_IMAGE_URL = "http://aries.ca/images/sample.png"


class TestDIDXResponseHandler(IsolatedAsyncioTestCase):
    def did_doc(self):
        doc = DIDDoc(did=TEST_DID)
        controller = TEST_DID
        ident = "1"
        pk_value = TEST_VERKEY
        pk = PublicKey(
            TEST_DID,
            ident,
            pk_value,
            PublicKeyType.ED25519_SIG_2018,
            controller,
            False,
        )
        doc.set(pk)
        recip_keys = [pk]
        router_keys = []
        service = Service(
            TEST_DID,
            "indy",
            "IndyAgent",
            recip_keys,
            router_keys,
            TEST_ENDPOINT,
        )
        doc.set(service)
        return doc

    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.ctx = RequestContext.test_context(self.profile)
        self.ctx.message_receipt = MessageReceipt()

        self.profile.context.injector.bind_instance(DIDMethods, DIDMethods())
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            self.did_info = await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
            )

            self.did_doc_attach = AttachDecorator.data_base64(self.did_doc().serialize())
            await self.did_doc_attach.data.sign(self.did_info.verkey, wallet)

            self.request = DIDXResponse(
                did=TEST_DID,
                did_doc_attach=self.did_doc_attach,
            )

    @mock.patch.object(test_module, "DIDXManager")
    async def test_called(self, mock_didx_mgr):
        mock_didx_mgr.return_value.accept_response = mock.CoroutineMock()
        self.ctx.message = DIDXResponse()
        handler_inst = test_module.DIDXResponseHandler()
        responder = MockResponder()
        await handler_inst.handle(self.ctx, responder)

        mock_didx_mgr.return_value.accept_response.assert_called_once_with(
            self.ctx.message, self.ctx.message_receipt
        )
        assert not responder.messages

    @mock.patch.object(test_module, "DIDXManager")
    async def test_called_auto_ping(self, mock_didx_mgr):
        self.ctx.update_settings({"auto_ping_connection": True})
        mock_didx_mgr.return_value.accept_response = mock.CoroutineMock()
        self.ctx.message = DIDXResponse()
        handler_inst = test_module.DIDXResponseHandler()
        responder = MockResponder()
        await handler_inst.handle(self.ctx, responder)

        mock_didx_mgr.return_value.accept_response.assert_called_once_with(
            self.ctx.message, self.ctx.message_receipt
        )
        messages = responder.messages
        assert len(messages) == 1
        result, _ = messages[0]
        assert isinstance(result, Ping)

    @mock.patch.object(test_module, "DIDXManager")
    @mock.patch.object(connection_target, "ConnectionTarget")
    async def test_problem_report(self, mock_conn_target, mock_didx_mgr):
        mock_didx_mgr.return_value.accept_response = mock.CoroutineMock(
            side_effect=DIDXManagerError(
                error_code=ProblemReportReason.RESPONSE_NOT_ACCEPTED.value
            )
        )
        mock_didx_mgr.return_value.manager_error_to_problem_report = mock.CoroutineMock(
            return_value=(
                DIDXProblemReport(
                    description={
                        "en": "test error",
                        "code": ProblemReportReason.RESPONSE_NOT_ACCEPTED.value,
                    }
                ),
                [mock_conn_target],
            )
        )
        self.ctx.message = DIDXResponse()
        handler_inst = test_module.DIDXResponseHandler()
        responder = MockResponder()
        await handler_inst.handle(self.ctx, responder)
        messages = responder.messages
        assert len(messages) == 1
        result, target = messages[0]
        assert (
            isinstance(result, DIDXProblemReport)
            and result.description
            and (
                result.description["code"]
                == ProblemReportReason.RESPONSE_NOT_ACCEPTED.value
            )
        )
        assert target == {"target_list": [mock_conn_target]}

    @mock.patch.object(test_module, "DIDXManager")
    @mock.patch.object(connection_target, "ConnectionTarget")
    async def test_problem_report_did_doc(
        self,
        mock_conn_target,
        mock_didx_mgr,
    ):
        mock_didx_mgr.return_value.accept_response = mock.CoroutineMock(
            side_effect=DIDXManagerError(
                error_code=ProblemReportReason.RESPONSE_NOT_ACCEPTED.value
            )
        )
        mock_didx_mgr.return_value.diddoc_connection_targets = mock.MagicMock(
            return_value=[mock_conn_target]
        )
        mock_didx_mgr.return_value.manager_error_to_problem_report = mock.CoroutineMock(
            return_value=(
                DIDXProblemReport(
                    description={
                        "en": "test error",
                        "code": ProblemReportReason.RESPONSE_NOT_ACCEPTED.value,
                    }
                ),
                [mock_conn_target],
            )
        )
        self.ctx.message = DIDXResponse(
            did=TEST_DID,
            did_doc_attach=self.did_doc_attach,
        )
        handler_inst = test_module.DIDXResponseHandler()
        responder = MockResponder()
        await handler_inst.handle(self.ctx, responder)
        messages = responder.messages
        assert len(messages) == 1
        result, target = messages[0]
        assert (
            isinstance(result, DIDXProblemReport)
            and result.description
            and (
                result.description["code"]
                == ProblemReportReason.RESPONSE_NOT_ACCEPTED.value
            )
        )
        assert target == {"target_list": [mock_conn_target]}

    @mock.patch.object(test_module, "DIDXManager")
    @mock.patch.object(connection_target, "ConnectionTarget")
    async def test_problem_report_did_doc_no_conn_target(
        self,
        mock_conn_target,
        mock_didx_mgr,
    ):
        mock_didx_mgr.return_value.accept_response = mock.CoroutineMock(
            side_effect=DIDXManagerError(
                error_code=ProblemReportReason.RESPONSE_NOT_ACCEPTED.value
            )
        )
        mock_didx_mgr.return_value.diddoc_connection_targets = mock.MagicMock(
            side_effect=DIDXManagerError("no target")
        )
        mock_didx_mgr.return_value.manager_error_to_problem_report = mock.CoroutineMock(
            return_value=(
                DIDXProblemReport(
                    description={
                        "en": "test error",
                        "code": ProblemReportReason.RESPONSE_NOT_ACCEPTED.value,
                    }
                ),
                None,
            )
        )
        self.ctx.message = DIDXResponse(
            did=TEST_DID,
            did_doc_attach=self.did_doc_attach,
        )
        handler_inst = test_module.DIDXResponseHandler()
        responder = MockResponder()
        await handler_inst.handle(self.ctx, responder)
        messages = responder.messages
        assert len(messages) == 0  # need target to add message



from unittest.mock import Mock

from acapy_agent.vc.ld_proofs.document_downloader import StaticCacheJsonLdDownloader


def test_load_cache_hit():
    downloader = Mock()
    context_loader = StaticCacheJsonLdDownloader(document_downloader=downloader)

    context_loader.load("https://www.w3.org/2018/credentials/v1")
    downloader.download.assert_not_called()


def test_load_cache_miss_triggers_download():
    downloader = Mock()
    downloader.download = Mock(return_value=(None, None))
    context_loader = StaticCacheJsonLdDownloader(document_downloader=downloader)

    context_loader.load("https://www.w3.org/2018/very_unlikely_name/v1")
    downloader.download.assert_called()



"""Test did resolver registry."""

import re
from typing import Pattern

import pytest
import pytest_asyncio
from pydid import DID, BasicDIDDocument, DIDDocument, VerificationMethod

from ...utils.testing import create_test_profile
from ..base import (
    BaseDIDResolver,
    DIDMethodNotSupported,
    DIDNotFound,
    ResolutionMetadata,
    ResolverError,
    ResolverType,
)
from ..did_resolver import DIDResolver
from . import DOC

TEST_DID0 = "did:sov:Kkyqu7CJFuQSvBp468uaDe"
TEST_DID1 = "did:btcr:8kyt-fzzq-qpqq-ljsc-5l"
TEST_DID2 = "did:ethr:mainnet:0xb9c5714089478a327f09197987f16f9e5d936e8a"
TEST_DID3 = "did:ion:EiDahaOGH-liLLdDtTxEAdc8i-cfCz-WUcQdRJheMVNn3A"
TEST_DID4 = "did:github:ghdid"
TEST_DID_5 = "did:key:zUC71nmwvy83x1UzNKbZbS7N9QZx8rqpQx3Ee3jGfKiEkZngTKzsRoqobX6wZdZF5F93pSGYYco3gpK9tc53ruWUo2tkBB9bxPCFBUjq2th8FbtT4xih6y6Q1K9EL4Th86NiCGT"

TEST_DIDS = [TEST_DID0, TEST_DID1, TEST_DID2, TEST_DID3, TEST_DID4, TEST_DID_5]

TEST_DID_METHOD0 = "sov"
TEST_DID_METHOD1 = "btcr"
TEST_DID_METHOD2 = "ethr"
TEST_DID_METHOD3 = "ion"
TEST_DID_METHOD4 = "github"
TEST_DID_METHOD5 = "key"

TEST_DID_METHODS = [
    TEST_DID_METHOD0,
    TEST_DID_METHOD1,
    TEST_DID_METHOD2,
    TEST_DID_METHOD3,
    TEST_DID_METHOD4,
    TEST_DID_METHOD5,
    "example",
]

TEST_METHOD_SPECIFIC_ID0 = "Kkyqu7CJFuQSvBp468uaDe"
TEST_METHOD_SPECIFIC_ID1 = "8kyt-fzzq-qpqq-ljsc-5l"
TEST_METHOD_SPECIFIC_ID2 = "mainnet:0xb9c5714089478a327f09197987f16f9e5d936e8a"
TEST_METHOD_SPECIFIC_ID3 = "EiDahaOGH-liLLdDtTxEAdc8i-cfCz-WUcQdRJheMVNn3A"
TEST_METHOD_SPECIFIC_ID4 = "ghdid"
TEST_METHOD_SPECIFIC_ID5 = "zUC71nmwvy83x1UzNKbZbS7N9QZx8rqpQx3Ee3jGfKiEkZngTKzsRoqobX6wZdZF5F93pSGYYco3gpK9tc53ruWUo2tkBB9bxPCFBUjq2th8FbtT4xih6y6Q1K9EL4Th86NiCGT"

TEST_METHOD_SPECIFIC_IDS = [
    TEST_METHOD_SPECIFIC_ID0,
    TEST_METHOD_SPECIFIC_ID1,
    TEST_METHOD_SPECIFIC_ID2,
    TEST_METHOD_SPECIFIC_ID3,
    TEST_METHOD_SPECIFIC_ID4,
    TEST_METHOD_SPECIFIC_ID5,
]


class MockResolver(BaseDIDResolver):
    def __init__(self, supported_methods, resolved=None, native: bool = False):
        super().__init__(ResolverType.NATIVE if native else ResolverType.NON_NATIVE)
        self._did_regex = re.compile(
            "^did:(?:{}):.*$".format("|".join(supported_methods))
        )
        self.resolved = resolved

    @property
    def supported_did_regex(self) -> Pattern:
        return self._did_regex

    async def setup(self, context):
        pass

    async def _resolve(self, profile, did, accept):
        if isinstance(self.resolved, Exception):
            raise self.resolved
        if isinstance(self.resolved, dict):
            return self.resolved
        return self.resolved.serialize()


@pytest.fixture
def resolver():
    did_resolver_registry = []
    for method in TEST_DID_METHODS:
        resolver = MockResolver([method], DIDDocument.deserialize(DOC))
        did_resolver_registry.append(resolver)
    return DIDResolver(did_resolver_registry)


@pytest_asyncio.fixture
async def profile():
    profile = await create_test_profile()
    yield profile


def test_create_resolver(resolver):
    assert len(resolver.resolvers) == len(TEST_DID_METHODS)


@pytest.mark.asyncio
@pytest.mark.parametrize("did, method", zip(TEST_DIDS, TEST_DID_METHODS))
async def test_match_did_to_resolver(profile, resolver, did, method):
    base_resolver, *_ = await resolver._match_did_to_resolver(profile, did)
    assert await base_resolver.supports(profile, did)


@pytest.mark.asyncio
async def test_match_did_to_resolver_x_not_supported(resolver):
    with pytest.raises(DIDMethodNotSupported):
        await resolver._match_did_to_resolver(
            profile, "did:cowsay:EiDahaOGH-liLLdDtTxEAdc8i-cfCz-WUcQdRJheMVNn3A"
        )


@pytest.mark.asyncio
async def test_match_did_to_resolver_native_priority(profile):
    native = MockResolver(["sov"], native=True)
    non_native = MockResolver(["sov"], native=False)
    registry = [non_native, native]
    resolver = DIDResolver(registry)
    assert [native, non_native] == await resolver._match_did_to_resolver(
        profile, TEST_DID0
    )


@pytest.mark.asyncio
async def test_match_did_to_resolver_registration_order(profile):
    native1 = MockResolver(["sov"], native=True)
    native2 = MockResolver(["sov"], native=True)
    non_native3 = MockResolver(["sov"], native=False)
    native4 = MockResolver(["sov"], native=True)
    registry = [native1, native2, non_native3, native4]
    resolver = DIDResolver(registry)
    assert [
        native1,
        native2,
        native4,
        non_native3,
    ] == await resolver._match_did_to_resolver(profile, TEST_DID0)


@pytest.mark.asyncio
async def test_dereference(resolver, profile):
    url = "did:example:1234abcd#4"
    expected: dict = DOC["verificationMethod"][0]
    actual: VerificationMethod = await resolver.dereference(profile, url)
    assert expected == actual.serialize()


@pytest.mark.asyncio
async def test_dereference_diddoc(resolver, profile):
    url = "did:example:1234abcd#4"
    doc = BasicDIDDocument(
        id="did:example:z6Mkmpe2DyE4NsDiAb58d75hpi1BjqbH6wYMschUkjWDEEuR"
    )
    result = await resolver.dereference(profile, url, document=doc)
    assert isinstance(result, VerificationMethod)
    assert result.id == url


@pytest.mark.asyncio
async def test_dereference_x(resolver, profile):
    url = "non-did"
    with pytest.raises(ResolverError):
        await resolver.dereference(profile, url)


@pytest.mark.asyncio
@pytest.mark.parametrize("did", TEST_DIDS)
async def test_resolve_with_metadata(resolver, profile, did):
    result = await resolver.resolve_with_metadata(profile, did)
    assert isinstance(result.did_document, dict)
    assert isinstance(result.metadata, ResolutionMetadata)


@pytest.mark.asyncio
@pytest.mark.parametrize("did", TEST_DIDS)
async def test_resolve(resolver, profile, did):
    doc = await resolver.resolve(profile, did)
    assert isinstance(doc, dict)


@pytest.mark.asyncio
@pytest.mark.parametrize("did", TEST_DIDS)
async def test_resolve_did(resolver, profile, did):
    doc = await resolver.resolve(profile, DID(did))
    assert isinstance(doc, dict)


@pytest.mark.asyncio
async def test_resolve_did_x_not_supported(resolver, profile):
    py_did = DID("did:cowsay:EiDahaOGH-liLLdDtTxEAdc8i-cfCz-WUcQdRJheMVNn3A")
    with pytest.raises(DIDMethodNotSupported):
        await resolver.resolve(profile, py_did)


@pytest.mark.asyncio
async def test_resolve_did_x_not_found(profile):
    py_did = DID("did:cowsay:EiDahaOGH-liLLdDtTxEAdc8i-cfCz-WUcQdRJheMVNn3A")
    cowsay_resolver_not_found = MockResolver(["cowsay"], resolved=DIDNotFound())
    resolver = DIDResolver([cowsay_resolver_not_found])
    with pytest.raises(DIDNotFound):
        await resolver.resolve(profile, py_did)



"""Connect invitation handler under RFC 23 (DID exchange)."""

from .....messaging.base_handler import BaseHandler, BaseResponder, RequestContext
from ....out_of_band.v1_0.messages.invitation import InvitationMessage
from ..messages.problem_report import DIDXProblemReport, ProblemReportReason


class InvitationHandler(BaseHandler):
    """Handler class for connection invitation message under RFC 23 (DID exchange)."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle connection invitation under RFC 23 (DID exchange).

        Args:
            context: Request context
            responder: Responder callback
        """

        self._logger.debug("InvitationHandler called with context %s", context)
        assert isinstance(context.message, InvitationMessage)

        report = DIDXProblemReport(
            description={
                "code": ProblemReportReason.INVITATION_NOT_ACCEPTED.value,
                "en": (
                    "Out-of-band invitations for DID exchange "
                    "cannot be submitted via agent messaging"
                ),
            }
        )
        report.assign_thread_from(context.message)
        # client likely needs to be using direct responses to receive the problem report
        await responder.send_reply(report)



"""keylist-query message used to request list of keys handled by mediator."""

from typing import Optional

from marshmallow import fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import KEYLIST_QUERY, PROTOCOL_PACKAGE
from .inner.keylist_query_paginate import KeylistQueryPaginate, KeylistQueryPaginateSchema

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.keylist_query_handler.KeylistQueryHandler"


class KeylistQuery(AgentMessage):
    """Class representing a keylist query message."""

    class Meta:
        """Metadata for a keylist query."""

        handler_class = HANDLER_CLASS
        message_type = KEYLIST_QUERY
        schema_class = "KeylistQuerySchema"

    def __init__(
        self,
        *,
        filter: Optional[dict] = None,
        paginate: Optional[KeylistQueryPaginate] = None,
        **kwargs,
    ):
        """Initialize keylist query object.

        Args:
            filter: Filter for query
            paginate: Pagination rules
            kwargs: Additional keyword arguments for the message
        """
        super().__init__(**kwargs)
        self.filter = filter
        self.paginate = paginate


class KeylistQuerySchema(AgentMessageSchema):
    """Keylist query schema class."""

    class Meta:
        """Keylist query schema metadata."""

        model_class = KeylistQuery

    filter = fields.Dict(
        required=False,
        metadata={"description": "Query dictionary object", "example": {"filter": {}}},
    )
    paginate = fields.Nested(
        KeylistQueryPaginateSchema(),
        required=False,
        metadata={"description": "Pagination info"},
    )
